;+
; NAME: H5REXPANDFILE
;
; PURPOSE:
;   This subroutine expands variables in a given HDF5 file and writes the result in another.
;
; CATEGORY:
;   Data processing.
;
; CALLING SEQUENCE:
;   H5EXPANDFILE, IDIR, ITYP, INUM, ODIR, OTYP, ONUM, ICHK
;
; INPUTS:
;   IDIR, ITYP, INUM:
;    the directory, type and number of input file/files.
;
;   ODIR, OTYP, ONUM:
;    the directory, type and number of output file/files.
;
;   ICHK:
;    number of the chunk
;
; OPTIONAL INPUT PARAMETERS:
;
; KEYWORD INPUT PARAMETERS:
;   PRINT:  Optional keyword increasing verbosity of this subroutine.
;
;
pro h5expandfile, idir, ityp, inum, odir, otyp, onum, ichk, print = p, _extra = extra

; generate input filename
;
  ifile = h5filename(idir,ityp,inum,ichk)

; check it the file is really written in HDF5 format
;
  if (not h5f_is_hdf5(ifile)) then begin
    message, 'The file ' + ifile + ' is not in HDF5 format or does not exist!'
  endif

; gather some information required for the processing procedure
;
  h5info, ifile, dims = dm, nguard = ng, pdims = pm, step = step, time = time, dt = dt   $
        , newdt = dtn, pcoords = pc, bext = bext, nfor = nf, efor = ef, cfor = cf        $
        , vfor = vf, sfor = sf, nseeds = nsds, seed = seed, nchunks = np, nvars = nvars  $
        , names = names, nrun = nrun, xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax $
        , zmin = zmin, zmax = zmax, inidt = dtini, mindt = dtmin, cfl = cfl, csnd = csnd $
        , gamma = gamma, pwfor = pwfor, fordt = dtfor, kf = kf, kl = kl, ku = ku, kc = kc

; prepare dimension and other information
;
  ii  = where(dm eq 1, ic)
  cm = dm - 2 * ng
  rm = 2 * cm
  fm = rm + 2 * ng
  pi = 2 * [ 1, 1, 1 ]
  if (ic ge 1) then begin
    cm[ii] = 1
    rm[ii] = 1
    fm[ii] = 1
    pi[ii] = 1
  endif
  fi = product(pi)

; print info
;
  if (keyword_set(p)) then begin
    print, 'guard cells                :', ng
    print, 'chunk dims                 :', cm
    print, 'chunk dims + guard cells   :', dm
    print, 'resized dims               :', rm
    print, 'resized dims + guard cells :', fm
  endif

; prepare subindices
;
  ib = intarr(3)
  ie = intarr(3)

  ib[*] = ng
  ie[*] = ng + fm - 1

; read coordinates
;
  xc = h5_read_coordinate(ifile, 'xc')
  yc = h5_read_coordinate(ifile, 'yc')
  zc = h5_read_coordinate(ifile, 'zc')

; expand coordinates
;
  xf = (interpolate_tvd(xc, /linear))[ib[0]:ie[0]]
  xi = xf + 0.5*(xf[1] - xf[0])
  dx = xf(2) - xf(1)
  yf = (interpolate_tvd(yc, /linear))[ib[1]:ie[1]]
  yi = yf + 0.5*(yf[1] - yf[0])
  dy = yf(2) - yf(1)
  if (ic ge 1) then begin
    zf = zc
    zi = zc
    dz = 1.0
  endif else begin
    zf = (interpolate_tvd(zc, /linear))[ib[2]:ie[2]]
    zi = zf + 0.5*(zf[1] - zf[0])
    dz = zf(2) - zf(1)
  endelse

; create output file and write coordinate attributes
;
  ofile = h5filename(odir,otyp,onum,ichk)

  fid = h5f_create(ofile)

; create group 'attributes'
;
  gid = h5g_create(fid, 'attributes')

; copy all attributes to this new file
;
  h5_create_attribute, gid, 'dims'   , fm
  if (np gt 1) then begin
    h5_create_attribute, gid, 'pdims'  , pm
    h5_create_attribute, gid, 'pcoords', pc
  endif
  h5_create_attribute, gid, 'bext'   , bext
  h5_create_attribute, gid, 'seed'   , seed
  h5_create_attribute, gid, 'step'   , step
  if (otyp eq 'r' or otyp eq 'd') then begin
    h5_create_attribute, gid, 'nrun'   , nrun
    h5_create_attribute, gid, 'ng'     , ng
  endif
  h5_create_attribute, gid, 'nseeds' , nsds
  if (np gt 1) then begin
    h5_create_attribute, gid, 'nproc'  , long(ichk)
    h5_create_attribute, gid, 'nprocs' , np
  endif
  if (nf gt 0) then begin
    h5_create_attribute, gid, 'nf'     , nf
    h5_create_attribute, gid, 'efor'   , ef * fi
    h5_create_attribute, gid, 'cfor'   , cf * fi
    h5_create_attribute, gid, 'vfor'   , vf * fi
    h5_create_attribute, gid, 'sfor'   , sf * fi
  endif
  h5_create_attribute, gid, 'xmin'   , xmin
  h5_create_attribute, gid, 'xmax'   , xmax
  h5_create_attribute, gid, 'ymin'   , ymin
  h5_create_attribute, gid, 'ymax'   , ymax
  h5_create_attribute, gid, 'zmin'   , zmin
  h5_create_attribute, gid, 'zmax'   , zmax
  h5_create_attribute, gid, 'dx'     , dx
  h5_create_attribute, gid, 'dy'     , dy
  h5_create_attribute, gid, 'dz'     , dz
  h5_create_attribute, gid, 'dxi'    , 1.0 / dx
  h5_create_attribute, gid, 'dyi'    , 1.0 / dy
  h5_create_attribute, gid, 'dzi'    , 1.0 / dz
  h5_create_attribute, gid, 'time'   , time
  h5_create_attribute, gid, 'dt'     , 0.5*dt
  h5_create_attribute, gid, 'dtn'    , 0.5*dtn
  h5_create_attribute, gid, 'dtini'  , dtini
  h5_create_attribute, gid, 'dtmin'  , dtmin
  h5_create_attribute, gid, 'cfl'    , cfl
  h5_create_attribute, gid, 'csnd'   , csnd
  h5_create_attribute, gid, 'gamma'  , gamma
  if (nf gt 0) then begin
    h5_create_attribute, gid, 'pwfor'  , pwfor
    h5_create_attribute, gid, 'dtfor'  , dtfor
    h5_create_attribute, gid, 'kf'     , kf
    h5_create_attribute, gid, 'kl'     , kl
    h5_create_attribute, gid, 'ku'     , ku
    h5_create_attribute, gid, 'kc'     , kc
  endif

; close group 'attributes'
;
  h5g_close, gid

; create and close group 'coordinates'
;
  gid = h5g_create(fid, 'coordinates')
  h5g_close, gid

; create and close group 'variables'
;
  gid = h5g_create(fid, 'variables')
  h5g_close, gid

; close file
;
  h5f_close, fid

; write coordinates to file
;
  h5_write_coordinate, ofile, 'xc', xf, _extra = extra
  h5_write_coordinate, ofile, 'yc', yf, _extra = extra
  h5_write_coordinate, ofile, 'zc', zf, _extra = extra
  h5_write_coordinate, ofile, 'xi', xi, _extra = extra
  h5_write_coordinate, ofile, 'yi', yi, _extra = extra
  h5_write_coordinate, ofile, 'zi', zi, _extra = extra

; expand and write veriables to file
;
  for n = 0, nvars - 1 do begin
    print, 'reading ', names[n], ' from file ', ifile

; read dataset
;
    iarr = h5_read_variable(ifile, names[n], coord = pc)

; expand dataset
;
    case names[n] of
    'potx': begin
            if (dm(2) gt 1) then begin
              temp = (expand_tvd(iarr, 1, /linear))[ib[0]:ie[0],*,*]
              temp = (expand_tvd(temp, 2, /center))[*,ib[1]:ie[1],*]
              oarr = (expand_tvd(temp, 3, /center))[*,*,ib[2]:ie[2]]
            endif else begin
              temp = (expand_tvd(iarr, 1, /linear))[ib[0]:ie[0],*]
              oarr = (expand_tvd(temp, 2, /center))[*,ib[1]:ie[1]]
            endelse
          end
    'poty': begin
            if (dm(2) gt 1) then begin
              temp = (expand_tvd(iarr, 1, /center))[ib[0]:ie[0],*,*]
              temp = (expand_tvd(temp, 2, /linear))[*,ib[1]:ie[1],*]
              oarr = (expand_tvd(temp, 3, /center))[*,*,ib[2]:ie[2]]
            endif else begin
              temp = (expand_tvd(iarr, 1, /center))[ib[0]:ie[0],*]
              oarr = (expand_tvd(temp, 2, /linear))[*,ib[1]:ie[1]]
            endelse
          end
    'potz': begin
            if (dm(2) gt 1) then begin
              temp = (expand_tvd(iarr, 1, /center))[ib[0]:ie[0],*,*]
              temp = (expand_tvd(temp, 2, /center))[*,ib[1]:ie[1],*]
              oarr = (expand_tvd(temp, 3, /linear))[*,*,ib[2]:ie[2]]
            endif else begin
              temp = (expand_tvd(iarr, 1, /center))[ib[0]:ie[0],*]
              oarr = (expand_tvd(temp, 2, /center))[*,ib[1]:ie[1]]
            endelse
          end
    else: begin
          if (dm(2) gt 1) then begin
            temp = (expand_tvd(iarr, 1))[ib[0]:ie[0],*,*]
            temp = (expand_tvd(temp, 2))[*,ib[1]:ie[1],*]
            oarr = (expand_tvd(temp, 3))[*,*,ib[2]:ie[2]]
          endif else begin
            temp = (expand_tvd(iarr, 1))[ib[0]:ie[0],*]
            oarr = (expand_tvd(temp, 2))[*,ib[1]:ie[1]]
          endelse
       end
    endcase

; write variable to the output file
;
     h5_write_variable, ofile, reform(names[n]), oarr, _extra = extra

  endfor
end
