;+
; NAME: H5REDISTRIBUTE
;
; PURPOSE:
;   This subroutine merges from or redistributes to the data stored in multiple HDF5 files.
;
; CATEGORY:
;   Data processing.
;
; CALLING SEQUENCE:
;   H5REDISTRIBUTE, IDIR, ITYP, INUM, ODIR, OTYP, ONUM, PDIM
;
; INPUTS:
;   IDIR, ITYP, INUM:
;    the directory, type and number of input file/files.
;
;   ODIR, OTYP, ONUM:
;    the directory, type and number of output file/files.
;
;   PDIM:
;    new division of data
;
; OPTIONAL INPUT PARAMETERS:
;
; KEYWORD INPUT PARAMETERS:
;   PRINT:  Optional keyword increasing verbosity of this subroutine.
;
;
pro h5redistribute, idir, ityp, inum, odir, otyp, onum, pdim, print = p, _extra = extra

  nr = long(product(pdim))

  if (nr le 0) then $
    message, 'New distribution should have positive non-zero dimensions!'

; generate input filename
;
  ichk = 0
  ifile = h5filename(idir,ityp,inum,ichk)

; check it the file is really written in HDF5 format
;
  if (not h5f_is_hdf5(ifile)) then begin
    message, 'The file ' + ifile + ' is not in HDF5 format or does not exist!'
  endif

; gather some information required for the processing procedure
;
  h5info, ifile, dims = dm, nguard = ng, pdims = pm, step = step, time = time, dt = dt   $
        , newdt = dtn, pcoords = pc, bext = bext, nfor = nf, efor = ef, cfor = cf        $
        , vfor = vf, sfor = sf, nseeds = nsds, seed = seed, nchunks = np, nvars = nvars  $
        , names = names, nrun = nrun, xmin = xmin, xmax = xmax, ymin = ymin, ymax = ymax $
        , zmin = zmin, zmax = zmax, inidt = dtini, mindt = dtmin, cfl = cfl, csnd = csnd $
        , dx = dx, dy = dy, dz = dz, gamma = gamma, pwfor = pwfor, fordt = dtfor         $
        , kf = kf, kl = kl, ku = ku, kc = kc

; prepare dimension and other information
;
  ii  = where(dm eq 1, ic)
  cm = dm - 2 * ng
  rm = pm * cm
  fm = rm + 2 * ng
  nm = rm / pdim
  qm = nm + 2 * ng
  if (ic ge 1) then begin
    cm[ii] = 1
    rm[ii] = 1
    fm[ii] = 1
    nm[ii] = 1
    qm[ii] = 1
  endif

; print info
;
  if (keyword_set(p)) then begin
    print, 'num. of chunks               :', np
    print, 'num. of new chunks           :', nr
    print, 'chunk distribution           :', pm
    print, 'new chunk distribution       :', long(pdim)
    print, 'guard cells                  :', ng
    print, 'chunk dims                   :', cm
    print, 'chunk dims + guard cells     :', dm
    print, 'total dims                   :', rm
    print, 'total dims + guard cells     :', fm
    print, 'new chunk dims               :', nm
    print, 'new chunk dims + guard cells :', qm
  endif

; expand seeds to new size
;
  if (2*nr ne nsds) then begin
    new_seed = rebin(seed, 2*nr)
    new_seed[ 0:nr-1] = seed[0]
    new_seed[nr:*   ] = seed[nsds-1]
    nseeds = 2*nr
    seed   = new_seed
  endif

; create new chunk files and write all attributes
;
  ochk = 0

; NOTE: do not change the order of loops; must be i, j, k for proper filename generation
;
  for i = 0, pdim[0] - 1 do $
    for j = 0, pdim[1] - 1 do $
      for k = 0, pdim[2] - 1 do begin

; create output file and write coordinate attributes
;
        ofile = h5filename(odir,otyp,onum,ochk)

        fid = h5f_create(ofile)

; create group 'attributes'
;
        gid = h5g_create(fid, 'attributes')

; copy all attributes to this new file
;
        h5_create_attribute, gid, 'dims'   , qm
        if (nr gt 1) then begin
          h5_create_attribute, gid, 'pdims'  , pdim
          h5_create_attribute, gid, 'pcoords', long([i, j, k])
        endif
        h5_create_attribute, gid, 'bext'   , bext
        h5_create_attribute, gid, 'seed'   , seed
        h5_create_attribute, gid, 'step'   , step
        if (otyp eq 'r' or otyp eq 'd') then begin
          h5_create_attribute, gid, 'nrun'   , nrun
          h5_create_attribute, gid, 'ng'     , ng
        endif
        h5_create_attribute, gid, 'nseeds' , nseeds
        if (nr gt 1) then begin
          h5_create_attribute, gid, 'nproc'  , ochk
          h5_create_attribute, gid, 'nprocs' , nr
        endif
        if (nf gt 0) then begin
          h5_create_attribute, gid, 'nf'     , nf
          h5_create_attribute, gid, 'efor'   , ef
          h5_create_attribute, gid, 'cfor'   , cf
          h5_create_attribute, gid, 'vfor'   , vf
          h5_create_attribute, gid, 'sfor'   , sf
        endif
        h5_create_attribute, gid, 'xmin'   , xmin
        h5_create_attribute, gid, 'xmax'   , xmax
        h5_create_attribute, gid, 'ymin'   , ymin
        h5_create_attribute, gid, 'ymax'   , ymax
        h5_create_attribute, gid, 'zmin'   , zmin
        h5_create_attribute, gid, 'zmax'   , zmax
        h5_create_attribute, gid, 'dx'     , dx
        h5_create_attribute, gid, 'dy'     , dy
        h5_create_attribute, gid, 'dz'     , dz
        h5_create_attribute, gid, 'dxi'    , 1.0 / dx
        h5_create_attribute, gid, 'dyi'    , 1.0 / dy
        h5_create_attribute, gid, 'dzi'    , 1.0 / dz
        h5_create_attribute, gid, 'time'   , time
        h5_create_attribute, gid, 'dt'     , dt
        h5_create_attribute, gid, 'dtn'    , dtn
        h5_create_attribute, gid, 'dtini'  , dtini
        h5_create_attribute, gid, 'dtmin'  , dtmin
        h5_create_attribute, gid, 'cfl'    , cfl
        h5_create_attribute, gid, 'csnd'   , csnd
        h5_create_attribute, gid, 'gamma'  , gamma
        if (nf gt 0) then begin
          h5_create_attribute, gid, 'pwfor'  , pwfor
          h5_create_attribute, gid, 'dtfor'  , dtfor
          h5_create_attribute, gid, 'kf'     , kf
          h5_create_attribute, gid, 'kl'     , kl
          h5_create_attribute, gid, 'ku'     , ku
          h5_create_attribute, gid, 'kc'     , kc
        endif

; close group 'attributes'
;
        h5g_close, gid

; create and close group 'coordinates'
;
        gid = h5g_create(fid, 'coordinates')
        h5g_close, gid

; create and close group 'variables'
;
        gid = h5g_create(fid, 'variables')
        h5g_close, gid

; close file
;
        h5f_close, fid

        ochk = ochk + 1
      endfor

; prepare subindices and arrays
;
  ib = intarr(3)
  ie = intarr(3)
  jb = intarr(3)
  je = intarr(3)

  arr = rebin([time], fm[0], fm[1], fm[2])
  xc  = rebin([time], fm[0])
  yc  = rebin([time], fm[1])
  zc  = rebin([time], fm[2])
  xi  = rebin([time], fm[0])
  yi  = rebin([time], fm[1])
  zi  = rebin([time], fm[2])

; for each variable merge dataset from old chunks and split into new distribution
;
  for n = 0, nvars - 1 do begin

    for ichk = 0, np - 1 do begin

; prepare filename
;
      ifile = h5filename(idir,ityp,inum,ichk)

      print, 'reading ', names[n], ' from file ', ifile

; read coordinates
;
      txc = h5_read_coordinate(ifile, 'xc')
      tyc = h5_read_coordinate(ifile, 'yc')
      tzc = h5_read_coordinate(ifile, 'zc')
      txi = h5_read_coordinate(ifile, 'xi')
      tyi = h5_read_coordinate(ifile, 'yi')
      tzi = h5_read_coordinate(ifile, 'zi')

; read dataset
;
      tmp = h5_read_variable(ifile, names[n], coord = pc)

; prepare indices
;
      for i = 0, 2 do begin

        if (dm[i] gt 1) then begin
          if (pc[i] eq 0) then begin
            ib[i] = 0
            jb[i] = 0
          endif else begin
            ib[i] = ng + pc[i] * cm[i]
            jb[i] = ng
          endelse

          if (pc[i] eq (pm[i] - 1)) then begin
            ie[i] = (pc[i] + 1) * cm[i] + 2 * ng - 1
            je[i] =               cm[i] + 2 * ng - 1
          endif else begin
            ie[i] = (pc[i] + 1) * cm[i] +     ng - 1
            je[i] =               cm[i] +     ng - 1
          endelse
        endif

      endfor

; insert dataset into variable array
;
      arr[ib[0]:ie[0],ib[1]:ie[1],ib[2]:ie[2]] = tmp[jb[0]:je[0],jb[1]:je[1],jb[2]:je[2]]
      xc [ib[0]:ie[0]] = txc[jb[0]:je[0]]
      yc [ib[1]:ie[1]] = tyc[jb[1]:je[1]]
      zc [ib[2]:ie[2]] = tzc[jb[2]:je[2]]
      xi [ib[0]:ie[0]] = txi[jb[0]:je[0]]
      yi [ib[1]:ie[1]] = tyi[jb[1]:je[1]]
      zi [ib[2]:ie[2]] = tzi[jb[2]:je[2]]

    endfor

; redistribute data and write to output files
;
  ochk = 0

; NOTE: do not change the order of loops; must be i, j, k for proper filename generation
;
  for i = 0, pdim[0] - 1 do begin
    ib[0] = i * nm[0]
    ie[0] = (i + 1) * nm[0] + 2 * ng - 1

    for j = 0, pdim[1] - 1 do begin
      ib[1] = j * nm[1]
      ie[1] = (j + 1) * nm[1] + 2 * ng - 1

      for k = 0, pdim[2] - 1 do begin
        ib[2] = k * nm[2]
        ie[2] = (k + 1) * nm[2] + 2 * ng - 1
        if (dm[2] le 1) then ie[2] = 0

; create output file and write coordinate attributes
;
        ofile = h5filename(odir,otyp,onum,ochk)

; write coordinates
;
        if (names[n] eq 'dens') then begin
          h5_write_coordinate, ofile, 'xc', xc[ib[0]:ie[0]], _extra = extra
          h5_write_coordinate, ofile, 'yc', yc[ib[1]:ie[1]], _extra = extra
          h5_write_coordinate, ofile, 'zc', zc[ib[2]:ie[2]], _extra = extra

          h5_write_coordinate, ofile, 'xi', xi[ib[0]:ie[0]], _extra = extra
          h5_write_coordinate, ofile, 'yi', yi[ib[1]:ie[1]], _extra = extra
          h5_write_coordinate, ofile, 'zi', zi[ib[2]:ie[2]], _extra = extra
        endif

; write variable
;
        tmp = arr[ib[0]:ie[0],ib[1]:ie[1],ib[2]:ie[2]]

        h5_write_variable, ofile, reform(names[n]), tmp, _extra = extra

        ochk = ochk + 1
      endfor
    endfor
  endfor

  endfor

end