;===============================================================================
;
;  function READHDF5_COORDINATE reads coordinates from HDF5 data cubes produced
;  by the Godunov code.
;
;-------------------------------------------------------------------------------
;
;  arguments:
;
;    file        - the file name of the HDF5 data snapshot
;    coordinate  - the coordinate name (xc, yc, zc, xi, yi, or zi)
;    restrict    - the restriction factor, meaning if the data should be
;                  rescaled down (only integer numbers larger than 1)
;
function readhdf5_coordinate, file, coordinate                                 $
                            , restrict = restrict, list = list, single = single

; initiate variables
;
  array = -1
  ncpus = 1

; check it the file exists
;
  if (not file_test(file, /regular, /read)) then begin

    print, file + ' does not exist!'

    return, array

  endif

; check it the file is really written in HDF format
;
  if (not h5f_is_hdf5(file)) then begin

    print, file + ' is not a HDF5 file!'

    return, array

  endif

; open a file
;
  fid = h5f_open(file)

; check if there were problems with opening the SD interface
;
  if (fid le 0) then begin

    print, 'Could not open ' + file

; close a file
;
    h5f_close, fid

; return -1
;
    return, -1

  endif

; everything goes fine so far, so proceed with the file
;
; if keyword /LIST is defined list all coordinates stored in the file
;
  if (keyword_set(list)) then begin

; open group 'coordinates'
;
    gid = h5g_open(fid, 'coordinates')

; get number of datasets
;
    ndatasets = h5g_get_num_objs(gid)

; print all dataset names
;
    print, 'Number of coordinates stored:', ndatasets
    names = strarr(ndatasets)
    for n = 0, ndatasets - 1 do begin
      nm = h5g_get_obj_name_by_idx(gid, n)
      names[n] = nm
    endfor
    print, 'Coordinate names            :', names

; close group 'coordinates'
;
    h5g_close, gid

; close a file
;
    h5f_close, fid

; return 0
;
    return, 0

  endif

; open group 'attributes'
;
  gid = h5g_open(fid, 'attributes')

; read attributes
;
  nprocs = 1

; get number of attributes
;
  nattributes = h5a_get_num_attrs(gid)

; read attributes
;
  for n = 0, nattributes - 1 do begin
    aid = h5a_open_idx(gid, n)
    name = h5a_get_name(aid)
    data = h5a_read(aid)
    case name of
      'nprocs' : nprocs = data[0]
    else :
    endcase
    h5a_close, aid
  endfor

; close group 'attributes'
;
  h5g_close, gid

; proceed if ncpus == 1 or the keyword /SINGLE is set
;
  if (nprocs eq 1 or keyword_set(single)) then begin ; data in a single file

; open group 'coordinates'
;
    gid = h5g_open(fid, 'coordinates')

; read coordinates
;
    did   = h5d_open(gid, coordinate)
    array = h5d_read(did)
    h5d_close, did

; close group 'coordinates'
;
    h5g_close, gid

; close the file
;
    h5f_close, fid

  endif else begin ; data in multiple files

; obtain information about the dimensions of the divisions
;
    pdims = [1, 1, 1]
    dims  = [1, 1, 1]
    ng    = 0
    nproc = 0

; open group 'attributes'
;
    gid = h5g_open(fid, 'attributes')

; read attributes
;
    for n = 0, nattributes - 1 do begin
      aid = h5a_open_idx(gid, n)
      name = h5a_get_name(aid)
      data = h5a_read(aid)
      case name of
        'pdims'  : pdims  = data
        'dims'   : dims   = data
        'ng'     : ng     = data[0]
        'nproc'  : nproc  = data[0]
      else :
      endcase
      h5a_close, aid
    endfor

; close group 'attributes'
;
    h5g_close, gid

; calculate  size of the full array
;
    ng = replicate(ng, n_elements(dims))
    ii = where(dims eq 1, c)
    if (c ge 1) then ng[ii] = 0
    sz = dims - 2 * ng

; create the array
;
      case coordinate of
        'xc': array = fltarr(pdims[0] * sz[0])
        'yc': array = fltarr(pdims[1] * sz[1])
        'zc': array = fltarr(pdims[2] * sz[2])
        'xi': array = fltarr(pdims[0] * sz[0])
        'yi': array = fltarr(pdims[1] * sz[1])
        'zi': array = fltarr(pdims[2] * sz[2])
      endcase

; close the file
;
    h5f_close, fid

; iterate over all files and read data
;
    pos = strpos(file, string(nproc, format = '(i5.5,".h5")'))

    for p = 0, nprocs-1 do begin

; prepare the filename
;
      strput, file, string(p, format = '(i5.5)'), pos

; open a file
;
      fid = h5f_open(file)

; open group 'attributes'
;
      gid = h5g_open(fid, 'attributes')

      for n = 0, nattributes - 1 do begin
        aid = h5a_open_idx(gid, n)
        name = h5a_get_name(aid)
        data = h5a_read(aid)
        case name of
          'pcoords': pcoords  = data
        else :
        endcase
        h5a_close, aid
      endfor

; close group 'attributes'
;
      h5g_close, gid

; open group 'coordinates'
;
      gid = h5g_open(fid, 'coordinates')

; read coordinates
;
      did = h5d_open(gid, coordinate)
      tmp = h5d_read(did)
      h5d_close, did

; close group 'coordinates'
;
      h5g_close, gid

; close a file
;
      h5f_close, fid

; insert sub-cube into global array
;
      case coordinate of
        'xc': begin
                ib  = ng[0]
                ie  = ib + sz[0] - 1

                tmp = temporary(tmp[ib:ie])

                ib  = pcoords[0] * sz[0]
                ie  = ib + sz[0] - 1

                array[ib:ie] = tmp
              end
        'yc': begin
                jb  = ng[1]
                je  = jb + sz[1] - 1

                tmp = temporary(tmp[jb:je])

                jb  = pcoords[1] * sz[1]
                je  = jb + sz[1] - 1

                array[jb:je] = tmp
              end
        'zc': begin
                kb  = ng[2]
                ke  = kb + sz[2] - 1

                tmp = temporary(tmp[kb:ke])

                kb  = pcoords[2] * sz[2]
                ke  = kb + sz[2] - 1

                array[kb:ke] = tmp
              end
        'xi': begin
                ib  = ng[0]
                ie  = ib + sz[0] - 1

                tmp = temporary(tmp[ib:ie])

                ib  = pcoords[0] * sz[0]
                ie  = ib + sz[0] - 1

                array[ib:ie] = tmp
              end
        'yi': begin
                jb  = ng[1]
                je  = jb + sz[1] - 1

                tmp = temporary(tmp[jb:je])

                jb  = pcoords[1] * sz[1]
                je  = jb + sz[1] - 1

                array[jb:je] = tmp
              end
        'zi': begin
                kb  = ng[2]
                ke  = kb + sz[2] - 1

                tmp = temporary(tmp[kb:ke])

                kb  = pcoords[2] * sz[2]
                ke  = kb + sz[2] - 1

                array[kb:ke] = tmp
              end
      endcase

    endfor

  endelse

; return the coordinate array
;
  return, array

end
