!!******************************************************************************
!!
!!  This file is part of the GODUNOV source code, a program to perform
!!  Newtonian or relativistic magnetohydrodynamical simulations.
!!
!!  Copyright (C) 2007-2013 Grzegorz Kowal <grzegorz@amuncode.org>
!!
!!  This program is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  This program is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  You should have received a copy of the GNU General Public License
!!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
!!
!!******************************************************************************
!!
!! module: DATAIO
!!
!!  This module handles data storage and job restart from restart files.
!!
!!******************************************************************************
!
module dataio

#ifdef PROFILE
! include external procedures
!
  use timers, only : set_timer, start_timer, stop_timer
#endif /* PROFILE */

! module variables are not implicit by default
!
  implicit none

#ifdef PROFILE
! timer indices
!
  integer           , save :: irs, iss
#endif /* PROFILE */

! all module parameters
!
  integer           , save :: nres    = -1
  character(len=255), save :: respath = "./"

! by default everything is private
!
  private

! declare public subroutines
!
  public :: initialize_dataio, store_snapshot                                  &
          , read_restart_snapshot, store_restart_snapshot

!- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
!
  contains
!
!===============================================================================
!
! subroutine INITIALIZE_DATAIO:
! ----------------------------
!
!   Subroutine initializes the module by setting its parameters.
!
!
!===============================================================================
!
  subroutine initialize_dataio()

! include external procedures and variables
!
    use parameters, only : get_parameter_integer, get_parameter_string

! local variables are not implicit by default
!
    implicit none
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! set timer descriptions
!
    call set_timer('snapshot reading', irs)
    call set_timer('snapshot storing', iss)
#endif /* PROFILE */

! read values of the module parameters
!
    call get_parameter_integer("nres"   , nres   )
    call get_parameter_string ("respath", respath)

!-------------------------------------------------------------------------------
!
  end subroutine initialize_dataio
!
!===============================================================================
!
! subroutine STORE_SNAPSHOT:
! -------------------------
!
!   Subroutine is a wrapper for data storing in the supported file format.
!
!
!===============================================================================
!
  subroutine store_snapshot(ftype, fnumber, nrun, iret)

! local variables are not implicit by default
!
    implicit none

! input and output arguments
!
    character, intent(in)    :: ftype
    integer  , intent(in)    :: fnumber
    integer  , intent(inout) :: nrun, iret
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! start the times accounting time spent for data storage
!
    call start_timer(iss)
#endif /* PROFILE */

#ifdef HDF5
! store a snapshot in the HDF5 format
!
    call store_snapshot_hdf5(ftype, fnumber, nrun, iret)
#endif /* HDF5 */

! check if storing data was successful
!
    if (iret > 0) iret = 125

#ifdef PROFILE
! stop the timer for data storage
!
    call stop_timer(iss)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine store_snapshot
!
!===============================================================================
!
! subroutine READ_RESTART_SNAPSHOT:
! --------------------------------
!
!   Subroutine is a wrapper for reading a problem setup from a restart snapshot.
!
!
!===============================================================================
!
  subroutine read_restart_snapshot(nrun, iret)

! local variables are not implicit by default
!
    implicit none

! output arguments
!
    integer, intent(inout) :: nrun, iret
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! start the times accounting time spent for data reading
!
    call start_timer(irs)
#endif /* PROFILE */

#ifdef HDF5
! store a snapshot in the HDF5 format
!
    call read_snapshot_hdf5(nrun, iret)
#endif /* HDF5 */

! check if reading restart files was successful
!
    if (iret > 0) iret = 124

#ifdef PROFILE
! stop the timer for data reading
!
    call stop_timer(irs)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine read_restart_snapshot
!
!===============================================================================
!
! subroutine STORE_RESTART_SNAPSHOT:
! ---------------------------------
!
!   Subroutine is a wrapper for storing restart snapshot in the supported
!   file format.
!
!
!===============================================================================
!
  subroutine store_restart_snapshot(nrun, iret)

! local variables are not implicit by default
!
    implicit none

! input and output arguments
!
    integer, intent(inout) :: nrun, iret
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! start the times accounting time spent for data storage
!
    call start_timer(iss)
#endif /* PROFILE */

#ifdef HDF5
! store a snapshot in the HDF5 format
!
    call store_snapshot_hdf5('r', nrun, nrun, iret)
#endif /* HDF5 */

! check if storing restart files was successful
!
    if (iret > 0) iret = 127

#ifdef PROFILE
! stop the timer for data storage
!
    call stop_timer(iss)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine store_restart_snapshot
#ifdef HDF5
!
!===============================================================================
!
! subroutine READ_SNAPSHOT_HDF5:
! -----------------------------
!
!   Subroutine reads all variables from restart files in order to continue
!   the system evolution.
!
!===============================================================================
!
  subroutine read_snapshot_hdf5(nrun, iret)

! include external procedures and variables
!
#ifdef FORCING
    use forcing   , only : nf
#ifdef INT_FOR
    use forcing   , only : efor, cfor, vfor, sfor, afrx, afry, afrz
#endif /* INT_FOR */
#endif /* FORCING */
    use hdf5      , only : H5F_ACC_RDONLY_F, H5T_NATIVE_INTEGER                &
                         , H5T_NATIVE_DOUBLE, H5T_NATIVE_REAL, hid_t, hsize_t
    use hdf5      , only : h5fis_hdf5_f, h5open_f, h5close_f                   &
                         , h5fopen_f, h5fclose_f, h5tcopy_f, h5tclose_f        &
                         , h5gopen_f, h5gclose_f                               &
                         , h5aget_num_attrs_f, h5aopen_idx_f                   &
                         , h5aget_name_f, h5aopen_name_f, h5aread_f            &
                         , h5aclose_f, h5dopen_f, h5dread_f, h5dclose_f
    use mesh      , only : im, jm, km
    use mesh      , only : step, time, dt, dtn
    use mpitools  , only : nproc, nprocs, pcoords
    use random    , only : nseeds, set_seeds
    use variables , only : nt, nn
    use variables , only : cvars, pvars
#ifdef BFIELD
    use variables , only : buni
#ifdef VPOT
    use variables , only : inx, iny, inz
    use variables , only : bi
#endif /* VPOT */
#endif /* BFIELD */
    use variables , only : u0, qp, qp_bnd
#ifdef HDF5EXT
    use variables , only : u1
#endif /* HDF5EXT */

! local variables are not implicit by default
!
    implicit none

! output arguments
!
    integer, intent(inout)         :: nrun, iret

! HDF5 variables
!
    integer(hid_t)                 :: fid, gid, aid, tid, did
    integer(hsize_t), dimension(3) :: dm
    integer(hsize_t), dimension(3) :: cm = (/ nt, 3, 2 /)
    integer(hsize_t), dimension(1) :: am = (/1/)
    integer(hsize_t)               :: alen = 16
    logical                        :: info

! local variables
!
    character(len=64)     :: fname
    character(len=16)     :: aname
    integer               :: n, nsds, nattrs
#ifdef MPI
    integer, dimension(3) :: coords
#endif /* MPI */

! allocatable vector to store random generator seeds
!
    integer(kind=4), dimension(:), allocatable :: seed
!
!-------------------------------------------------------------------------------
!
! prepare the restart file number for a given path and file number
!
    write (fname, "(a,'r',i6.6,'_',i5.5,'.h5')") trim(respath), nres, nproc

! check if the file exists
!
    inquire(file = fname, exist = info)
    if (.not. info) then
      write (*,'(a)') 'ERROR: ' // trim(fname) //                              &
                      ' does not exist! Please check your setup.'
      iret = 120
      return
    end if

! check if the file content is stored in the HDF5 format
!
    call h5fis_hdf5_f(fname, info, iret)
    if (.not. info) then
      write (*,'(a)') 'ERROR: ' // trim(fname) //                              &
                      ' is not an HDF5 file! Please check your setup.'
      iret = 121
      return
    end if

! prepare variable dimensions
!
    dm(:) = (/ im, jm, km /)

! initial the size of seed vector
!
    nsds  = 0

! initialize the HDF5 Fortran interface
!
    call h5open_f(iret)

! prepare datatypes
!
    call h5tcopy_f(H5T_NATIVE_DOUBLE, tid, iret)

! open the HDF5 file read only
!
    call h5fopen_f(fname, H5F_ACC_RDONLY_F, fid, iret)

! open the attribute group
!
    call h5gopen_f(fid, 'attributes', gid, iret)

! obtain the number of attributes
!
    call h5aget_num_attrs_f(gid, nattrs, iret)

! iterate over all attributes
!
    do n = 0, nattrs - 1

! open the current attribute
!
      call h5aopen_idx_f(gid, n, aid, iret)

! get its name
!
      call h5aget_name_f(aid, alen, aname, iret)

! restore the attributes which are crucial for the problem restart
!
      select case(trim(aname))

      case('dt')
        am(1) = 1
        call h5aread_f(aid, tid, dt, am, iret)

      case('dtn')
        am(1) = 1
        call h5aread_f(aid, tid, dtn, am, iret)

      case('nrun')
        am(1) = 1
        call h5aread_f(aid, H5T_NATIVE_INTEGER, nrun, am, iret)
        nrun = nrun + 1

      case('nseeds')
        am(1) = 1
        call h5aread_f(aid, H5T_NATIVE_INTEGER, nsds, am, iret)

      case('step')
        am(1) = 1
        call h5aread_f(aid, H5T_NATIVE_INTEGER, step, am, iret)

      case('time')
        am(1) = 1
        call h5aread_f(aid, tid, time, am, iret)

      case('qp_bnd')
        call h5aread_f(aid, H5T_NATIVE_DOUBLE, qp_bnd(:,:,:), cm, iret)

#ifdef BFIELD
      case('buni')
        am(1) = 3
        call h5aread_f(aid, H5T_NATIVE_DOUBLE, buni(:), am, iret)
#endif /* BFIELD */

#ifdef FORCING
      case('nf')
        am(1) = 1
        call h5aread_f(aid, H5T_NATIVE_INTEGER, nf, am, iret)
#ifdef INT_FOR

      case('efor')
        am(1) = 1
        call h5aread_f(aid, H5T_NATIVE_DOUBLE, efor, am, iret)

      case('cfor')
        am(1) = 1
        call h5aread_f(aid, H5T_NATIVE_DOUBLE, cfor, am, iret)

      case('vfor')
        am(1) = 1
        call h5aread_f(aid, H5T_NATIVE_DOUBLE, vfor, am, iret)

      case('sfor')
        am(1) = 1
        call h5aread_f(aid, H5T_NATIVE_DOUBLE, sfor, am, iret)

      case('afrx')
        am(1) = 1
        call h5aread_f(aid, H5T_NATIVE_DOUBLE, afrx, am, iret)

      case('afry')
        am(1) = 1
        call h5aread_f(aid, H5T_NATIVE_DOUBLE, afry, am, iret)

      case('afrz')
        am(1) = 1
        call h5aread_f(aid, H5T_NATIVE_DOUBLE, afrz, am, iret)
#endif /* INT_FOR */
#endif /* FORCING */

#ifdef MPI
      case('pcoords')
        am(1) = 3
        coords(:) = 0
        call h5aread_f(aid, H5T_NATIVE_INTEGER, coords, am, iret)

        if (sum(abs(coords-pcoords)) .ne. 0) then
          write (*,'(a)') 'ERROR   : coords of file and process are different!'
          write (*,'(a,i4,a,3i3,a,3i3)') 'ERROR   : ncpu = ', nproc            &
                            , ' file = ', coords, ' proc = ', pcoords
          iret = 122
          return
        end if
#endif /* MPI */

      case default
      end select

! close the current attribute
!
      call h5aclose_f(aid, iret)

    end do ! n = 0, nattrs - 1

! restore the state of random number generator
!
    if (nsds .eq. nseeds .and. nseeds .gt. 0) then

      am(1) = nseeds
      allocate(seed(0:nseeds-1))
      call h5aopen_name_f(gid, 'seed', aid, iret)
      call h5aread_f(aid, H5T_NATIVE_INTEGER, seed, am, iret)
      call h5aclose_f(aid, iret)
      call set_seeds(nseeds, seed)
      deallocate(seed)

    end if

! close the attribute group
!
    call h5gclose_f(gid, iret)

! open the variable group
!
    call h5gopen_f(fid, 'variables', gid, iret)

! restore the current state conserved variables
!
    do n = 1, nt

      call h5dopen_f(gid, 'u0:' // cvars(n), did, iret)
      call h5dread_f(did, tid, u0(n,:,:,:), dm(1:3), iret)
      call h5dclose_f(did, iret)

    end do

#ifdef HDF5EXT
! restore the previous state conserved variables
!
    do n = 1, nt

      call h5dopen_f(gid, 'u1:' // cvars(n), did, iret)
      call h5dread_f(did, tid, u1(n,:,:,:), dm(1:3), iret)
      call h5dclose_f(did, iret)

    end do
#endif /* HDF5EXT */

! restore the current state primitive variables
!
    do n = 1, nn

      call h5dopen_f(gid, 'qp:' // pvars(n), did, iret)
      call h5dread_f(did, tid, qp(n,:,:,:), dm(1:3), iret)
      call h5dclose_f(did, iret)

    end do

#ifdef VPOT
! restore the current state face-centered magnetic field components
!
    call h5dopen_f(gid, 'bi:magx', did, iret)
    call h5dread_f(did, tid, bi(inx,:,:,:), dm(1:3), iret)
    call h5dclose_f(did, iret)

    call h5dopen_f(gid, 'bi:magy', did, iret)
    call h5dread_f(did, tid, bi(iny,:,:,:), dm(1:3), iret)
    call h5dclose_f(did, iret)

    call h5dopen_f(gid, 'bi:magz', did, iret)
    call h5dread_f(did, tid, bi(inz,:,:,:), dm(1:3), iret)
    call h5dclose_f(did, iret)
#endif /* VPOT */

! close the variable group
!
    call h5gclose_f(gid, iret)

! terminate access to the file
!
    call h5fclose_f(fid, iret)

! release the datatype
!
    call h5tclose_f(tid, iret)

! close the HDF5 Fortran interface
!
    call h5close_f(iret)

!-------------------------------------------------------------------------------
!
  end subroutine read_snapshot_hdf5
!
!===============================================================================
!
! subroutine STORE_SNAPSHOT_HDF5:
! ------------------------------
!
!   Subroutine writes all data in Hierarchical Data Format (HDF) v5 for
!   specification and libraries see http://hdf.ncsa.uiuc.edu.
!
!   Arguments:
!
!     ftype   - determines which data should be written:
!       'p' -> default, zone-center interpolated primitive variables only
!              (only domain and single precision),
!       'r' -> full actual U0 and U1 if necessary for restart dumps
!              (domain with boundaries and current precision),
!       'f' -> full dumps (as in 'p' + all additional variables like energies,
!              div B, etc.; without boundaries and with single precision),
!
!     fnumber - file number of the current dump;
!
!===============================================================================
!
  subroutine store_snapshot_hdf5(ftype, fnumber, nrun, iret)

! include external procedures and variables
!
#ifdef FORCING
    use forcing   , only : nf, pwfor, dtfor, kl, kf, ku, kc
#ifdef INT_FOR
    use forcing   , only : efor, cfor, vfor, sfor, afrx, afry, afrz
#endif /* INT_FOR */
#endif /* FORCING */
    use hdf5
    use mesh      , only : step, time, dt, dtn, dti
    use mesh      , only : ng, in, jn, kn                                      &
                         , xmin, xmax, ymin, ymax, zmin, zmax
    use mesh      , only : im, jm, km, ib, ie, jb, je, kb, ke                  &
                         , dx, dy, dz, dxi, dyi, dzi, dxh, dyh, dzh
    use mesh      , only : xc, yc, zc, xi, yi, zi
#ifdef CD
    use mesh      , only : dxih, dyih, dzih
#endif /* CD */
    use mpitools  , only : nproc, nprocs, pdims, pcoords
    use operators , only : divergence
    use random    , only : nseeds, get_seeds
    use variables , only : nt, nn
    use variables , only : cvars, pvars
    use variables , only : qp, u0, du
    use variables , only : buni, qp_bnd
#ifdef HDF5EXT
    use variables , only : u1
#endif /* HDF5EXT */
    use variables , only : idn, imx, imy, imz
    use variables , only : ivx, ivy, ivz
#ifdef ADI
    use variables , only : ipr, ien
#endif /* ADI */
#ifdef BFIELD
    use variables , only : ibx, iby, ibz
    use variables , only : inx, iny, inz
#ifdef VPOT
    use variables , only : iax, iay, iaz
    use variables , only : bi
#endif /* VPOT */
#endif /* BFIELD */
#ifdef SHOCK_DETECTION
    use variables , only : sh
#endif /* SHOCK_DETECTION */

! local variables are not implicit by default
!
    implicit none

! input arguments
!
    character, intent(in)    :: ftype
    integer  , intent(in)    :: fnumber
    integer  , intent(inout) :: nrun, iret

! HDF5 variables
!
    character(len=64)              :: fname
    integer(hid_t)                 :: fid, gid, aid, tid, sid, did, pid
    integer(hsize_t), dimension(3) :: dm
    integer(hsize_t), dimension(3) :: cm = (/ nt, 3, 2 /)
    integer(hsize_t), dimension(1) :: am
    integer         , dimension(3) :: dims

! local variables
!
    integer                        :: n, i, j, k, im1, jm1, km1, ip1, jp1, kp1

! allocatable variables
!
    integer(kind=4), dimension(:)    , allocatable :: seed
!
!-------------------------------------------------------------------------------
!
! prepare dimensions depending on the snapshot type
!
    if (ftype .eq. 'r') then
      dims(1) = im
      dims(2) = jm
      dims(3) = km
    else
      dims(1) = in
      dims(2) = jn
      dims(3) = kn
    end if
    dm(:) = dims(:)

! prepare the file name
!
    write(fname,'(a1,i6.6,"_",i5.5,".h5")') ftype, fnumber, nproc

! initialize the HDF5 Fortran interface
!
    call h5open_f(iret)

! create or overwrite the HDF5 file
!
    call h5fcreate_f(fname, H5F_ACC_TRUNC_F, fid, iret)

! prepare a datatype for stored variables
!
    call h5tcopy_f(H5T_NATIVE_DOUBLE, tid, iret)

! prepare properties / compression
!
    call h5pcreate_f(H5P_DATASET_CREATE_F, pid, iret)
    if (ftype .eq. 'r') then
#ifdef HDF5GZIP
      call h5pset_chunk_f(pid, 3, dm(:), iret)
      call h5pset_deflate_f(pid, 9, iret)
#endif /* HDF5GZIP */
#ifdef HDF5SZIP
      call h5pset_chunk_f(pid, 3, dm(:), iret)
      call h5pset_szip_f(pid, H5_SZIP_NN_OM_F, 8, iret)
#endif /* HDF5SZIP */
    else
      call h5pset_chunk_f(pid, 3, dm(:), iret)
      call h5pset_deflate_f(pid, 9, iret)
    end if

! create a group to store attributes
!
    call h5gcreate_f(fid, 'attributes', gid, iret)

! store vector parameters
!
    am(1) = 3
    call h5screate_simple_f(1, am, sid, iret)

    call h5acreate_f(gid, 'dims' , H5T_NATIVE_INTEGER, sid, aid, iret)
    call h5awrite_f(aid, H5T_NATIVE_INTEGER, dims, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'pdims', H5T_NATIVE_INTEGER, sid, aid, iret)
    call h5awrite_f(aid, H5T_NATIVE_INTEGER, pdims, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'pcoords', H5T_NATIVE_INTEGER, sid, aid, iret)
    call h5awrite_f(aid, H5T_NATIVE_INTEGER, pcoords, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'buni', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, buni, am, iret)
    call h5aclose_f(aid, iret)

    call h5sclose_f(sid, iret)

    if (ftype .eq. 'r') then

! store boundary values
!
      call h5screate_simple_f(3, cm, sid, iret)

      call h5acreate_f(gid, 'qp_bnd', tid, sid, aid, iret)
      call h5awrite_f(aid, tid, qp_bnd, cm, iret)
      call h5aclose_f(aid, iret)

      call h5sclose_f(sid, iret)

    end if

! store random number generator parameters
!
    allocate(seed(0:nseeds-1))
    call get_seeds(seed(:))
    am(1) = nseeds
    call h5screate_simple_f(1, am, sid, iret)
    call h5acreate_f(gid, 'seed', H5T_NATIVE_INTEGER, sid, aid, iret)
    call h5awrite_f(aid, H5T_NATIVE_INTEGER, seed, am, iret)
    call h5aclose_f(aid, iret)
    call h5sclose_f(sid, iret)
    deallocate(seed)

! store scalar parameters
!
    am(1) = 1
    call h5screate_simple_f(1, am, sid, iret)

    call h5acreate_f(gid, 'step', H5T_NATIVE_INTEGER, sid, aid, iret)
    call h5awrite_f(aid, H5T_NATIVE_INTEGER, step, am, iret)
    call h5aclose_f(aid, iret)

    if (ftype .eq. 'r') then

      call h5acreate_f(gid, 'nrun', H5T_NATIVE_INTEGER, sid, aid, iret)
      call h5awrite_f(aid, H5T_NATIVE_INTEGER, nrun, am, iret)
      call h5aclose_f(aid, iret)

      call h5acreate_f(gid, 'ng', H5T_NATIVE_INTEGER, sid, aid, iret)
      call h5awrite_f(aid, H5T_NATIVE_INTEGER, ng, am, iret)
      call h5aclose_f(aid, iret)

    end if

    call h5acreate_f(gid, 'nseeds', H5T_NATIVE_INTEGER, sid, aid, iret)
    call h5awrite_f(aid, H5T_NATIVE_INTEGER, nseeds, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'nproc', H5T_NATIVE_INTEGER, sid, aid, iret)
    call h5awrite_f(aid, H5T_NATIVE_INTEGER, nproc, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'nprocs', H5T_NATIVE_INTEGER, sid, aid, iret)
    call h5awrite_f(aid, H5T_NATIVE_INTEGER, nprocs, am, iret)
    call h5aclose_f(aid, iret)

#ifdef FORCING
    call h5acreate_f(gid, 'nf', H5T_NATIVE_INTEGER, sid, aid, iret)
    call h5awrite_f(aid, H5T_NATIVE_INTEGER, nf, am, iret)
    call h5aclose_f(aid, iret)

#ifdef INT_FOR
    call h5acreate_f(gid, 'efor', H5T_NATIVE_DOUBLE, sid, aid, iret)
    call h5awrite_f(aid, H5T_NATIVE_DOUBLE, efor, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'cfor', H5T_NATIVE_DOUBLE, sid, aid, iret)
    call h5awrite_f(aid, H5T_NATIVE_DOUBLE, cfor, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'vfor', H5T_NATIVE_DOUBLE, sid, aid, iret)
    call h5awrite_f(aid, H5T_NATIVE_DOUBLE, vfor, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'sfor', H5T_NATIVE_DOUBLE, sid, aid, iret)
    call h5awrite_f(aid, H5T_NATIVE_DOUBLE, sfor, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'afrx', H5T_NATIVE_DOUBLE, sid, aid, iret)
    call h5awrite_f(aid, H5T_NATIVE_DOUBLE, afrx, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'afry', H5T_NATIVE_DOUBLE, sid, aid, iret)
    call h5awrite_f(aid, H5T_NATIVE_DOUBLE, afry, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'afrz', H5T_NATIVE_DOUBLE, sid, aid, iret)
    call h5awrite_f(aid, H5T_NATIVE_DOUBLE, afrz, am, iret)
    call h5aclose_f(aid, iret)
#endif /* INT_FOR */
#endif /* FORCING */

    call h5acreate_f(gid, 'xmin', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, xmin, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'xmax', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, xmax, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'ymin', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, ymin, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'ymax', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, ymax, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'zmin', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, zmin, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'zmax', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, zmax, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'dx', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, dx, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'dy', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, dy, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'dz', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, dz, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'dxi', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, dxi, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'dyi', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, dyi, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'dzi', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, dzi, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'time', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, time, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'dt', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, dt, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'dtn', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, dtn, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'dti', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, dti, am, iret)
    call h5aclose_f(aid, iret)

#ifdef FORCING
    call h5acreate_f(gid, 'pwfor', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, pwfor, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'dtfor', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, dtfor, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'kf', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, kf, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'kl', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, kl, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'ku', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, ku, am, iret)
    call h5aclose_f(aid, iret)

    call h5acreate_f(gid, 'kc', tid, sid, aid, iret)
    call h5awrite_f(aid, tid, kc, am, iret)
    call h5aclose_f(aid, iret)
#endif /* FORCING */

    call h5sclose_f(sid, iret)

! close the attribute group
!
    call h5gclose_f(gid, iret)

! create a group to store coordinates
!
    call h5gcreate_f(fid, 'coordinates', gid, iret)

    if (ftype .eq. 'r') then

! store the X coordinate
!
      call h5screate_simple_f(1, dm(1), sid, iret)
      call h5dcreate_f(gid, 'xc', tid, sid, did, iret)
      call h5dwrite_f(did, tid, xc(:), dm(1:1), iret, sid)
      call h5dclose_f(did, iret)
      call h5sclose_f(sid, iret)

! store the Y coordinate
!
      call h5screate_simple_f(1, dm(2), sid, iret)
      call h5dcreate_f(gid, 'yc', tid, sid, did, iret)
      call h5dwrite_f(did, tid, yc(:), dm(2:2), iret, sid)
      call h5dclose_f(did, iret)
      call h5sclose_f(sid, iret)

! store the Z coordinate
!
      call h5screate_simple_f(1, dm(3), sid, iret)
      call h5dcreate_f(gid, 'zc', tid, sid, did, iret)
      call h5dwrite_f(did, tid, zc(:), dm(3:3), iret, sid)
      call h5dclose_f(did, iret)
      call h5sclose_f(sid, iret)

! store the X interface coordinates
!
      call h5screate_simple_f(1, dm(1), sid, iret)
      call h5dcreate_f(gid, 'xi', tid, sid, did, iret)
      call h5dwrite_f(did, tid, xi(:), dm(1:1), iret, sid)
      call h5dclose_f(did, iret)
      call h5sclose_f(sid, iret)

! store the Y interface coordinates
!
      call h5screate_simple_f(1, dm(2), sid, iret)
      call h5dcreate_f(gid, 'yi', tid, sid, did, iret)
      call h5dwrite_f(did, tid, yi(:), dm(2:2), iret, sid)
      call h5dclose_f(did, iret)
      call h5sclose_f(sid, iret)

! store the Z interface coordinates
!
      call h5screate_simple_f(1, dm(3), sid, iret)
      call h5dcreate_f(gid, 'zi', tid, sid, did, iret)
      call h5dwrite_f(did, tid, zi(:), dm(3:3), iret, sid)
      call h5dclose_f(did, iret)
      call h5sclose_f(sid, iret)

    else

! store the X coordinate
!
      call h5screate_simple_f(1, dm(1), sid, iret)
      call h5dcreate_f(gid, 'xc', tid, sid, did, iret)
      call h5dwrite_f(did, tid, xc(ib:ie), dm(1:1), iret, sid)
      call h5dclose_f(did, iret)
      call h5sclose_f(sid, iret)

! store the Y coordinate
!
      call h5screate_simple_f(1, dm(2), sid, iret)
      call h5dcreate_f(gid, 'yc', tid, sid, did, iret)
      call h5dwrite_f(did, tid, yc(jb:je), dm(2:2), iret, sid)
      call h5dclose_f(did, iret)
      call h5sclose_f(sid, iret)

! store the Z coordinate
!
      call h5screate_simple_f(1, dm(3), sid, iret)
      call h5dcreate_f(gid, 'zc', tid, sid, did, iret)
      call h5dwrite_f(did, tid, zc(kb:ke), dm(3:3), iret, sid)
      call h5dclose_f(did, iret)
      call h5sclose_f(sid, iret)

! store the X interface coordinates
!
      call h5screate_simple_f(1, dm(1), sid, iret)
      call h5dcreate_f(gid, 'xi', tid, sid, did, iret)
      call h5dwrite_f(did, tid, xi(ib:ie), dm(1:1), iret, sid)
      call h5dclose_f(did, iret)
      call h5sclose_f(sid, iret)

! store the Y interface coordinates
!
      call h5screate_simple_f(1, dm(2), sid, iret)
      call h5dcreate_f(gid, 'yi', tid, sid, did, iret)
      call h5dwrite_f(did, tid, yi(jb:je), dm(2:2), iret, sid)
      call h5dclose_f(did, iret)
      call h5sclose_f(sid, iret)

! store the Z interface coordinates
!
      call h5screate_simple_f(1, dm(3), sid, iret)
      call h5dcreate_f(gid, 'zi', tid, sid, did, iret)
      call h5dwrite_f(did, tid, zi(kb:ke), dm(3:3), iret, sid)
      call h5dclose_f(did, iret)
      call h5sclose_f(sid, iret)

    end if

! close the attribute group
!
    call h5gclose_f(gid, iret)

! create a group to store variables
!
    call h5gcreate_f(fid, 'variables', gid, iret)

! create a dataspace for variables
!
    call h5screate_simple_f(3, dm(1:3), sid, iret)

! write the restart file
!
    if (ftype .eq. 'r') then

! store the current state conserved variables
!
      do n = 1, nt

        call h5dcreate_f(gid, 'u0:' // cvars(n), tid, sid, did, iret, pid)
        call h5dwrite_f(did, tid, u0(n,:,:,:), dm(1:3), iret, sid)
        call h5dclose_f(did, iret)

      end do

#ifdef HDF5EXT
! store the previous state conserved variables
!
      do n = 1, nt

        call h5dcreate_f(gid, 'u1:' // cvars(n), tid, sid, did, iret, pid)
        call h5dwrite_f(did, tid, u1(n,:,:,:), dm(1:3), iret, sid)
        call h5dclose_f(did, iret)

      end do
#endif /* HDF5EXT */

! store the current state primitive variables
!
      do n = 1, nn

        call h5dcreate_f(gid, 'qp:' // pvars(n), tid, sid, did, iret, pid)
        call h5dwrite_f(did, tid, qp(n,:,:,:), dm(1:3), iret, sid)
        call h5dclose_f(did, iret)

      end do

#ifdef VPOT
! store the current state face-centered magnetic field components
!
      call h5dcreate_f(gid, 'bi:magx', tid, sid, did, iret, pid)
      call h5dwrite_f(did, tid, bi(inx,:,:,:), dm(1:3), iret, sid)
      call h5dclose_f(did, iret)

      call h5dcreate_f(gid, 'bi:magy', tid, sid, did, iret, pid)
      call h5dwrite_f(did, tid, bi(iny,:,:,:), dm(1:3), iret, sid)
      call h5dclose_f(did, iret)

      call h5dcreate_f(gid, 'bi:magz', tid, sid, did, iret, pid)
      call h5dwrite_f(did, tid, bi(inz,:,:,:), dm(1:3), iret, sid)
      call h5dclose_f(did, iret)
#endif /* VPOT */

    end if

! write primitive variables and all additional variables if 'f' is the file type
!
    if (ftype .eq. 'p') then

! store all primitive variables
!
      do n = 1, nt
        call h5dcreate_f(gid, pvars(n), H5T_NATIVE_REAL, sid, did, iret, pid)
        call h5dwrite_f(did, H5T_NATIVE_REAL, real(qp(n,ib:ie,jb:je,kb:ke), 4) &
                                                         , dm(1:3), iret, sid)
        call h5dclose_f(did, iret)
      end do

    end if

! store primitive and additional variables if desired
!
    if (ftype .eq. 'f') then

      do n = 1, nn
        call h5dcreate_f(gid, pvars(n), H5T_NATIVE_REAL, sid, did, iret, pid)
        call h5dwrite_f(did, H5T_NATIVE_REAL, real(qp(n,ib:ie,jb:je,kb:ke), 4) &
                                                         , dm(1:3), iret, sid)
        call h5dclose_f(did, iret)
      end do

#ifdef ADI
! store the total energy
!
      call h5dcreate_f(gid, 'ener', H5T_NATIVE_REAL, sid, did, iret, pid)
      call h5dwrite_f(did, H5T_NATIVE_REAL, real(u0(ien,ib:ie,jb:je,kb:ke), 4) &
                                                         , dm(1:3), iret, sid)
      call h5dclose_f(did, iret)
#endif /* ADI */

! calculate the kinetic energy
!
      du(1,:,:,:) = 0.5d+00 * qp(idn,:,:,:)                                    &
                               * sum(qp(ivx:ivz,:,:,:) * qp(ivx:ivz,:,:,:), 1)

! store the kinetic energy
!
      call h5dcreate_f(gid, 'ekin', H5T_NATIVE_REAL, sid, did, iret, pid)
      call h5dwrite_f(did, H5T_NATIVE_REAL, real(du(1,ib:ie,jb:je,kb:ke), 4)   &
                                                         , dm(1:3), iret, sid)
      call h5dclose_f(did, iret)

! calculate the cell centered divergence of velocity
!
      call divergence(im, jm, km, qp(ivx:ivz,:,:,:), du(1,:,:,:), 'c')

! store the divergence of velocity
!
      call h5dcreate_f(gid, 'divv', H5T_NATIVE_REAL, sid, did, iret, pid)
      call h5dwrite_f(did, H5T_NATIVE_REAL, real(du(1,ib:ie,jb:je,kb:ke), 4)   &
                                                         , dm(1:3), iret, sid)
      call h5dclose_f(did, iret)

#ifdef BFIELD
! calculate the magnetic energy
!
      du(1,:,:,:) = 0.5d+00 * sum(qp(ibx:ibz,:,:,:) * qp(ibx:ibz,:,:,:), 1)

! store the magnetic energy
!
      call h5dcreate_f(gid, 'emag', H5T_NATIVE_REAL, sid, did, iret, pid)
      call h5dwrite_f(did, H5T_NATIVE_REAL, real(du(1,ib:ie,jb:je,kb:ke), 4)   &
                                                         , dm(1:3), iret, sid)
      call h5dclose_f(did, iret)

! store the vector potential
!
#ifdef CD
      call h5dcreate_f(gid, 'potx', H5T_NATIVE_REAL, sid, did, iret, pid)
      call h5dwrite_f(did, H5T_NATIVE_REAL, real(u0(iax,ib:ie,jb:je,kb:ke), 4) &
                                                         , dm(1:3), iret, sid)
      call h5dclose_f(did, iret)

      call h5dcreate_f(gid, 'poty', H5T_NATIVE_REAL, sid, did, iret, pid)
      call h5dwrite_f(did, H5T_NATIVE_REAL, real(u0(iay,ib:ie,jb:je,kb:ke), 4) &
                                                         , dm(1:3), iret, sid)
      call h5dclose_f(did, iret)

      call h5dcreate_f(gid, 'potz', H5T_NATIVE_REAL, sid, did, iret, pid)
      call h5dwrite_f(did, H5T_NATIVE_REAL, real(u0(iaz,ib:ie,jb:je,kb:ke), 4) &
                                                         , dm(1:3), iret, sid)
      call h5dclose_f(did, iret)
#endif /* CD */

#ifdef CT
      do k = kb, ke
#ifdef R3D
        km1 = k - 1
#endif /* R3D */
        do j = jb, je
          jm1 = j - 1
          do i = ib, ie
            im1 = i - 1
#ifdef R3D
            du(1,i,j,k) = (u0(iax,i  ,j  ,k  ) + u0(iax,i  ,jm1,k  )           &
                         + u0(iax,i  ,j  ,km1) + u0(iax,i  ,jm1,km1)) / 4.0d0
            du(2,i,j,k) = (u0(iay,i  ,j  ,k  ) + u0(iay,im1,j  ,k  )           &
                         + u0(iay,i  ,j  ,km1) + u0(iay,im1,j  ,km1)) / 4.0d0
#else /* R3D */
            du(1,i,j,k) = 0.5 * (u0(iax,i  ,j  ,k  ) + u0(iax,i  ,jm1,k  ))
            du(2,i,j,k) = 0.5 * (u0(iay,i  ,j  ,k  ) + u0(iay,im1,j  ,k  ))
#endif /* R3D */
            du(3,i,j,k) = (u0(iaz,i  ,j  ,k  ) + u0(iaz,im1,j  ,k  )           &
                         + u0(iaz,i  ,jm1,k  ) + u0(iaz,im1,jm1,k  )) / 4.0d0
          end do
        end do
      end do

      call h5dcreate_f(gid, 'potx', H5T_NATIVE_REAL, sid, did, iret, pid)
      call h5dwrite_f(did, H5T_NATIVE_REAL, real(du(1,ib:ie,jb:je,kb:ke), 4)   &
                                                         , dm(1:3), iret, sid)
      call h5dclose_f(did, iret)

      call h5dcreate_f(gid, 'poty', H5T_NATIVE_REAL, sid, did, iret, pid)
      call h5dwrite_f(did, H5T_NATIVE_REAL, real(du(2,ib:ie,jb:je,kb:ke), 4)   &
                                                         , dm(1:3), iret, sid)
      call h5dclose_f(did, iret)

      call h5dcreate_f(gid, 'potz', H5T_NATIVE_REAL, sid, did, iret, pid)
      call h5dwrite_f(did, H5T_NATIVE_REAL, real(du(3,ib:ie,jb:je,kb:ke), 4)   &
                                                         , dm(1:3), iret, sid)
      call h5dclose_f(did, iret)
#endif /* CT */

! calculate the cell centered divergence of magnetic field
!
#ifdef CT
      call divergence(im, jm, km, bi(inx:inz,:,:,:), du(1,:,:,:), 'f')
#endif /* CT */
#if defined CD || defined GLM
      call divergence(im, jm, km, qp(ibx:ibz,:,:,:), du(1,:,:,:), 'c')
#endif /* CD | GLM */

! store the divergence of magnetic field
!
      call h5dcreate_f(gid, 'divb', H5T_NATIVE_REAL, sid, did, iret, pid)
      call h5dwrite_f(did, H5T_NATIVE_REAL, real(du(1,ib:ie,jb:je,kb:ke), 4)   &
                                                         , dm(1:3), iret, sid)
      call h5dclose_f(did, iret)
#endif /* BFIELD */

#ifdef SHOCK_DETECTION
! convert shock locations to a real array
!
      du(1,:,:,:) = 0.0d+00
      where (sh .eqv. .true.)
        du(1,:,:,:) = 1.0d+00
      end where

! store shock locations
!
      call h5dcreate_f(gid, 'shck', H5T_NATIVE_REAL, sid, did, iret, pid)
      call h5dwrite_f(did, H5T_NATIVE_REAL, real(du(1,ib:ie,jb:je,kb:ke), 4)   &
                                                         , dm(1:3), iret, sid)
      call h5dclose_f(did, iret)
#endif /* SHOCK_DETECTION */

    end if

! release and terminate access to the dataspace
!
    call h5sclose_f(sid, iret)

! close the variable group
!
    call h5gclose_f(gid, iret)

! release properties
!
    call h5pclose_f(pid, iret)

! release datatypes
!
    call h5tclose_f(tid, iret)

! terminate access to the file
!
    call h5fclose_f(fid, iret)

! close the HDF5 Fortran interface
!
    call h5close_f(iret)

!-------------------------------------------------------------------------------
!
  end subroutine store_snapshot_hdf5
#endif /* HDF5 */

!===============================================================================
!
end module dataio
