!!******************************************************************************
!!
!!  This file is part of the GODUNOV source code, a program to perform
!!  Newtonian or relativistic magnetohydrodynamical simulations.
!!
!!  Copyright (C) 2007-2013 Grzegorz Kowal <grzegorz@amuncode.org>
!!
!!  This program is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  This program is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  You should have received a copy of the GNU General Public License
!!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
!!
!!******************************************************************************
!!
!! module: OPERATORS
!!
!!  This module provides differential operators and subroutines to calculate
!!  magnetic field from vector potential.
!!
!!******************************************************************************
!
module operators

#ifdef PROFILE
! include external procedures
!
  use timers, only : set_timer, start_timer, stop_timer
#endif /* PROFILE */

! module variables are not implicit by default
!
  implicit none

#ifdef PROFILE
! timer indices
!
  integer     , save :: idv, icr, ilp, iin, iab, iaj
#endif /* PROFILE */

! module parameters
!
  integer, parameter :: inx = 1, iny = 2, inz = 3

! by default everything is private
!
  private

! declare public subroutines
!
  public :: initialize_operators, divergence, curl, atob, atoj

!- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
!
  contains
!
!===============================================================================
!
! subroutine INITIALIZE_OPERATORS:
! -------------------------------
!
!   Subroutine initializes module OPERATORS.
!
!===============================================================================
!
  subroutine initialize_operators()

! local variables are not implicit by default
!
    implicit none
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! set timer descriptions
!
    call set_timer('divergence operator' , idv)
    call set_timer('curl operator'       , icr)
    call set_timer('laplace operator'    , ilp)
    call set_timer('interpolate operator', iin)
    call set_timer('A to B operator'     , iab)
    call set_timer('A to J operator'     , iaj)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine initialize_operators
!
!===============================================================================
!
! subroutine DERIVATIVE_1ST:
! -------------------------
!
!   Subroutine calculates the first order derivative of the input vector.  This
!   subroutine is used in many operators.
!
!   Arguments:
!
!     n - the direction of derivative
!     u - the input vector
!     v - the first derivative of the input vector
!     t - the location of the input field ('c' for central, 'f' for face)
!
!===============================================================================
!
  subroutine derivative_1st(n, im, jm, km, u, v, t)

! include external procedures and variables
!
    use mesh, only : dxi , dyi , dzi
    use mesh, only : dxih, dyih, dzih

! local variables are not implicit by default
!
    implicit none

! subroutine arguments
!
    integer                   , intent(in)  :: n, im, jm, km
    real, dimension(im,jm,km) , intent(in)  :: u
    real, dimension(im,jm,km) , intent(out) :: v
    character(len=1)          , intent(in)  :: t

! local variables
!
#ifdef DETECT_DISCONTINUITY
    integer         :: i  , j  , k
    integer         :: im1, jm1, km1
    integer         :: ip1, jp1, kp1
#ifndef TVD
    integer         :: im2, jm2, km2
    integer         :: ip2, jp2, kp2
#endif /* !TVD */
    real            :: dm, dp, dl, dh, du
#else /* DETECT_DISCONTINUITY */
    integer         :: im1, im2, jm1, jm2, km1, km2
#ifndef TVD
    integer         :: im3, im4, jm3, jm4, km3, km4
#endif /* !TVD */
#endif /* DETECT_DISCONTINUITY */

! coefficients for the 4th order derivative estimation
!
#ifndef TVD
    real, parameter :: cl = 2.0d+00 / 3.0d+00, cr = 1.0d+00 / 12.0d+00
    real, parameter :: fl = 9.0d+00 / 8.0d+00, fr = 1.0d+00 / 24.0d+00
#endif /* !TVD */
!
!-------------------------------------------------------------------------------
!
#ifdef DETECT_DISCONTINUITY
! face derivative (the value is stored in the front face)
!
    if (t == 'f') then

! select the direction
!
      select case(n)

! derivative along the X direction
!
      case(1)

        do i = 1, im

! prepare index shifts
!
          im1 = max( 1, i - 1)
#ifndef TVD
          im2 = max( 1, i - 2)
          ip1 = min(im, i + 1)
#endif /* !TVD */

! iterate over the perpendicular plane
!
          do k = 1, km
            do j = 1, jm

! calculate first order derivative using shorter stencil
!
              dl = u(i  ,j,k) - u(im1,j,k)

#ifdef TVD
! return the first order derivative
!
              v(i,j,k) = dl * dxi
#else /* TVD */
! calculate first order derivative using larger stencil
!
              dh = u(ip1,j,k) - u(im2,j,k)

! calculate the first derivative with higher order
!
              du = fl * dl - fr * dh

! calculate monotonicity test differences
!
              dm = 0.5d+00 * (u(i  ,j,k) - u(im2,j,k)) - du
              dp = 0.5d+00 * (u(ip1,j,k) - u(im1,j,k)) - du

! return the high order estimate of the first derivative
!
              if (dm * dp > 0.0d+00) then
                v(i,j,k) = dl * dxi
              else
                v(i,j,k) = du * dxi
              end if
#endif /* TVD */

            end do
          end do

        end do

! derivative along the Y direction
!
      case(2)

        do j = 1, jm

! prepare index shifts
!
          jm1 = max( 1, j - 1)
#ifndef TVD
          jm2 = max( 1, j - 2)
          jp1 = min(jm, j + 1)
#endif /* !TVD */

! iterate over the perpendicular plane
!
          do k = 1, km
            do i = 1, im

! calculate first order derivative using shorter stencil
!
              dl = u(i,j  ,k) - u(i,jm1,k)

#ifdef TVD
! return the first order derivative
!
              v(i,j,k) = dl * dyi
#else /* TVD */
! calculate first order derivative using larger stencil
!
              dh = u(i,jp1,k) - u(i,jm2,k)

! calculate the first derivative with higher order
!
              du = fl * dl - fr * dh

! calculate monotonicity test differences
!
              dm = 0.5d+00 * (u(i,j  ,k) - u(i,jm2,k)) - du
              dp = 0.5d+00 * (u(i,jp1,k) - u(i,jm1,k)) - du

! return the high order estimate of the first derivative
!
              if (dm * dp > 0.0d+00) then
                v(i,j,k) = dl * dyi
              else
                v(i,j,k) = du * dyi
              end if
#endif /* TVD */

            end do
          end do

        end do

! derivative along the Z direction
!
      case(3)

#ifdef R3D
        do k = 1, km

! prepare index shifts
!
          km1 = max( 1, k - 1)
#ifndef TVD
          km2 = max( 1, k - 2)
          kp1 = min(km, k + 1)
#endif /* !TVD */

! iterate over the perpendicular plane
!
          do j = 1, jm
            do i = 1, im

! calculate first order derivative using shorter stencil
!
              dl = u(i,j,k  ) - u(i,j,km1)

#ifdef TVD
! return the first order derivative
!
              v(i,j,k) = dl * dzi
#else /* TVD */
! calculate first order derivative using larger stencil
!
              dh = u(i,j,kp1) - u(i,j,km2)

! calculate the first derivative with higher order
!
              du = fl * dl - fr * dh

! calculate monotonicity test differences
!
              dm = 0.5d+00 * (u(i,j,k  ) - u(i,j,km2)) - du
              dp = 0.5d+00 * (u(i,j,kp1) - u(i,j,km1)) - du

! return the high order estimate of the first derivative
!
              if (dm * dp > 0.0d+00) then
                v(i,j,k) = dl * dzi
              else
                v(i,j,k) = du * dzi
              end if
#endif /* TVD */

            end do
          end do

        end do
#else /* R3D */
        v(:,:,:    ) = 0.0d+00
#endif /* R3D */

      end select

! face derivative (the value is stored in the back face)
!
    else if (t == 'b') then

! select the direction
!
      select case(n)

! derivative along the X direction
!
      case(1)

        do i = 1, im

! prepare index shifts
!
          ip1 = min(im, i + 1)
#ifndef TVD
          ip2 = min(im, i + 2)
          im1 = max( 1, i - 1)
#endif /* !TVD */

! iterate over the perpendicular plane
!
          do k = 1, km
            do j = 1, jm

! calculate first order derivative using shorter stencil
!
              dl = u(ip1,j,k) - u(i  ,j,k)

#ifdef TVD
! return the first order derivative
!
              v(i,j,k) = dl * dxi
#else /* TVD */
! calculate first order derivative using larger stencil
!
              dh = u(ip2,j,k) - u(im1,j,k)

! calculate the first derivative with higher order
!
              du = fl * dl - fr * dh

! calculate monotonicity test differences
!
              dm = 0.5d+00 * (u(ip1,j,k) - u(im1,j,k)) - du
              dp = 0.5d+00 * (u(i  ,j,k) - u(ip2,j,k)) - du

! return the high order estimate of the first derivative
!
              if (dm * dp > 0.0d+00) then
                v(i,j,k) = dl * dxi
              else
                v(i,j,k) = du * dxi
              end if
#endif /* TVD */

            end do
          end do

        end do

! derivative along the Y direction
!
      case(2)

        do j = 1, jm

! prepare index shifts
!
          jp1 = min(jm, j + 1)
#ifndef TVD
          jp2 = min(jm, j + 2)
          jm1 = max( 1, j - 1)
#endif /* !TVD */

! iterate over the perpendicular plane
!
          do k = 1, km
            do i = 1, im

! calculate first order derivative using shorter stencil
!
              dl = u(i,jp1,k) - u(i,j  ,k)

#ifdef TVD
! return the first order derivative
!
              v(i,j,k) = dl * dyi
#else /* TVD */
! calculate first order derivative using larger stencil
!
              dh = u(i,jp2,k) - u(i,jm1,k)

! calculate the first derivative with higher order
!
              du = fl * dl - fr * dh

! calculate monotonicity test differences
!
              dm = 0.5d+00 * (u(i,jp1,k) - u(i,jm1,k)) - du
              dp = 0.5d+00 * (u(i,jp2,k) - u(i,j  ,k)) - du

! return the high order estimate of the first derivative
!
              if (dm * dp > 0.0d+00) then
                v(i,j,k) = dl * dyi
              else
                v(i,j,k) = du * dyi
              end if
#endif /* TVD */

            end do
          end do

        end do

! derivative along the Z direction
!
      case(3)

#ifdef R3D
        do k = 1, km

! prepare index shifts
!
          kp1 = min(km, k + 1)
#ifndef TVD
          kp2 = min(km, k + 2)
          km1 = max( 1, k - 1)
#endif /* !TVD */

! iterate over the perpendicular plane
!
          do j = 1, jm
            do i = 1, im

! calculate first order derivative using shorter stencil
!
              dl = u(i,j,kp1) - u(i,j,k  )

#ifdef TVD
! return the first order derivative
!
              v(i,j,k) = dl * dzi
#else /* TVD */
! calculate first order derivative using larger stencil
!
              dh = u(i,j,kp2) - u(i,j,km1)

! calculate the first derivative with higher order
!
              du = fl * dl - fr * dh

! calculate monotonicity test differences
!
              dm = 0.5d+00 * (u(i,j,kp1) - u(i,j,km1)) - du
              dp = 0.5d+00 * (u(i,j,kp2) - u(i,j,k  )) - du

! return the high order estimate of the first derivative
!
              if (dm * dp > 0.0d+00) then
                v(i,j,k) = dl * dzi
              else
                v(i,j,k) = du * dzi
              end if
#endif /* TVD */

            end do
          end do

        end do
#else /* R3D */
        v(:,:,:    ) = 0.0d+00
#endif /* R3D */

      end select

    else

! select the direction
!
      select case(n)

! derivative along the X direction
!
      case(1)

        do i = 1, im

! prepare index shifts
!
          im1 = max( 1, i - 1)
          ip1 = min(im, i + 1)
#ifndef TVD
          im2 = max( 1, i - 2)
          ip2 = min(im, i + 2)
#endif /* !TVD */

! iterate over the perpendicular plane
!
          do k = 1, km
            do j = 1, jm

! calculate first order derivative using shorter stencil
!
              dl = u(ip1,j,k) - u(im1,j,k)

#ifdef TVD
! return the first order derivative
!
              v(i,j,k) = dl * dxih
#else /* TVD */
! calculate first order derivative using larger stencil
!
              dh = u(ip2,j,k) - u(im2,j,k)

! calculate the first derivative with higher order
!
              du = cl * dl - cr * dh

! calculate monotonicity test differences
!
              dm = (u(i  ,j,k) - u(im1,j,k)) - du
              dp = (u(ip1,j,k) - u(i  ,j,k)) - du

! return the high order estimate of the first derivative
!
              if (dm * dp > 0.0d+00) then
                v(i,j,k) = dl * dxih
              else
                v(i,j,k) = du * dxi
              end if
#endif /* TVD */

            end do
          end do

        end do

! derivative along the Y direction
!
      case(2)

        do j = 1, jm

! prepare index shifts
!
          jm1 = max( 1, j - 1)
          jp1 = min(jm, j + 1)
#ifndef TVD
          jm2 = max( 1, j - 2)
          jp2 = min(jm, j + 2)
#endif /* !TVD */

! iterate over the perpendicular plane
!
          do k = 1, km
            do i = 1, im

! calculate first order derivative using shorter stencil
!
              dl = u(i,jp1,k) - u(i,jm1,k)

#ifdef TVD
! return the first order derivative
!
              v(i,j,k) = dl * dyih
#else /* TVD */
! calculate first order derivative using larger stencil
!
              dh = u(i,jp2,k) - u(i,jm2,k)

! calculate the first derivative with higher order
!
              du = cl * dl - cr * dh

! calculate monotonicity test differences
!
              dm = (u(i,j  ,k) - u(i,jm1,k)) - du
              dp = (u(i,jp1,k) - u(i,j  ,k)) - du

! return the high order estimate of the first derivative
!
              if (dm * dp > 0.0d+00) then
                v(i,j,k) = dl * dyih
              else
                v(i,j,k) = du * dyi
              end if
#endif /* TVD */

            end do
          end do

        end do

! derivative along the Z direction
!
      case(3)

#ifdef R3D
        do k = 1, km

! prepare index shifts
!
          km1 = max( 1, k - 1)
          kp1 = min(km, k + 1)
#ifndef TVD
          km2 = max( 1, k - 2)
          kp2 = min(km, k + 2)
#endif /* !TVD */

! iterate over the perpendicular plane
!
          do j = 1, jm
            do i = 1, im

! calculate first order derivative using shorter stencil
!
              dl = u(i,j,kp1) - u(i,j,km1)

#ifdef TVD
! return the first order derivative
!
              v(i,j,k) = dl * dzih
#else /* TVD */
! calculate first order derivative using larger stencil
!
              dh = u(i,j,kp2) - u(i,j,km2)

! calculate the first derivative with higher order
!
              du = cl * dl - cr * dh

! calculate monotonicity test differences
!
              dm = (u(i,j,k  ) - u(i,j,km1)) - du
              dp = (u(i,j,kp1) - u(i,j,k  )) - du

! return the high order estimate of the first derivative
!
              if (dm * dp > 0.0d+00) then
                v(i,j,k) = dl * dzih
              else
                v(i,j,k) = du * dzi
              end if
#endif /* TVD */

            end do
          end do

        end do
#else /* R3D */
        v(:,:,:    ) = 0.0d+00
#endif /* R3D */

      end select

    end if
#else /* DETECT_DISCONTINUITY */
! prepare index limits during the first execution
!
    im1 = im - 1
    im2 = im - 2
#ifndef TVD
    im3 = im - 3
    im4 = im - 4
#endif /* !TVD */
    jm1 = jm - 1
    jm2 = jm - 2
#ifndef TVD
    jm3 = jm - 3
    jm4 = jm - 4
#endif /* !TVD */
#ifdef R3D
    km1 = km - 1
    km2 = km - 2
#ifndef TVD
    km3 = km - 3
    km4 = km - 4
#endif /* !TVD */
#endif /* R3D */

! face derivative (the value is stored in the front face)
!
    if (t == 'f') then

! select the direction
!
      select case(n)

! derivative along the X direction
!
      case(1)

#ifdef TVD
        v(2:im ,:,:) = dxi  *       (u(2:im ,:,:) - u(1:im1,:,:))
#else /* TVD */
        v(3:im1,:,:) = dxi  * (fl * (u(3:im1,:,:) - u(2:im2,:,:))              &
                             - fr * (u(4:im ,:,:) - u(1:im3,:,:)))
        v(  im ,:,:) = dxi  *       (u(  im ,:,:) - u(  im1,:,:))
        v(2    ,:,:) = dxi  *       (u(2    ,:,:) - u(1    ,:,:))
#endif /* TVD */
        v(1    ,:,:) = dxi  *       (u(2    ,:,:) - u(1    ,:,:))

! derivative along the Y direction
!
      case(2)

#ifdef TVD
        v(:,2:jm ,:) = dyi  *       (u(:,2:jm ,:) - u(:,1:jm1,:))
#else /* TVD */
        v(:,3:jm1,:) = dyi  * (fl * (u(:,3:jm1,:) - u(:,2:jm2,:))              &
                             - fr * (u(:,4:jm ,:) - u(:,1:jm3,:)))
        v(:,  jm ,:) = dyi  *       (u(:,  jm ,:) - u(:,  jm1,:))
        v(:,2    ,:) = dyi  *       (u(:,2    ,:) - u(:,1    ,:))
#endif /* TVD */
        v(:,1    ,:) = dyi  *       (u(:,2    ,:) - u(:,1    ,:))

! derivative along the Z direction
!
      case(3)

#ifdef R3D
#ifdef TVD
        v(:,:,2:km ) = dzi  *       (u(:,:,2:km ) - u(:,:,1:km1))
#else /* TVD */
        v(:,:,3:km1) = dzi  * (fl * (u(:,:,3:km1) - u(:,:,2:km2))              &
                             - fr * (u(:,:,4:km ) - u(:,:,1:km3)))
        v(:,:,  km ) = dzi  *       (u(:,:,  km ) - u(:,:,  km1))
        v(:,:,2    ) = dzi  *       (u(:,:,2    ) - u(:,:,1    ))
#endif /* TVD */
        v(:,:,1    ) = dzi  *       (u(:,:,2    ) - u(:,:,1    ))
#else /* R3D */
        v(:,:,:    ) = 0.0d0
#endif /* R3D */

      end select

! face derivative (the value is stored in the back face)
!
    else if (t == 'b') then

! select the direction
!
      select case(n)

! derivative along the X direction
!
      case(1)

#ifdef TVD
        v(1:im1,:,:) = dxi  *       (u(2:im ,:,:) - u(1:im1,:,:))
#else /* TVD */
        v(2:im2,:,:) = dxi  * (fl * (u(3:im1,:,:) - u(2:im2,:,:))              &
                             - fr * (u(4:im ,:,:) - u(1:im3,:,:)))
        v(  im1,:,:) = dxi  *       (u(  im ,:,:) - u(  im1,:,:))
        v(1    ,:,:) = dxi  *       (u(2    ,:,:) - u(1    ,:,:))
#endif /* TVD */
        v(  im ,:,:) = dxi  *       (u(  im ,:,:) - u(  im1,:,:))

! derivative along the Y direction
!
      case(2)

#ifdef TVD
        v(:,1:jm1,:) = dyi  *       (u(:,2:jm ,:) - u(:,1:jm1,:))
#else /* TVD */
        v(:,2:jm2,:) = dyi  * (fl * (u(:,3:jm1,:) - u(:,2:jm2,:))              &
                             - fr * (u(:,4:jm ,:) - u(:,1:jm3,:)))
        v(:,  jm1,:) = dyi  *       (u(:,  jm ,:) - u(:,  jm1,:))
        v(:,1    ,:) = dyi  *       (u(:,2    ,:) - u(:,1    ,:))
#endif /* TVD */
        v(:,  jm ,:) = dyi  *       (u(:,  jm ,:) - u(:,  jm1,:))

! derivative along the Z direction
!
      case(3)

#ifdef R3D
#ifdef TVD
        v(:,:,1:km1) = dzi  *       (u(:,:,2:km ) - u(:,:,1:km1))
#else /* TVD */
        v(:,:,2:km2) = dzi  * (fl * (u(:,:,3:km1) - u(:,:,2:km2))              &
                             - fr * (u(:,:,4:km ) - u(:,:,1:km3)))
        v(:,:,  km1) = dzi  *       (u(:,:,  km ) - u(:,:,  km1))
        v(:,:,1    ) = dzi  *       (u(:,:,2    ) - u(:,:,1    ))
#endif /* TVD */
        v(:,:,  km ) = dzi  *       (u(:,:,  km ) - u(:,:,  km1))
#else /* R3D */
        v(:,:,:    ) = 0.0d0
#endif /* R3D */

      end select

    else

! select the direction
!
      select case(n)

! derivative along the X direction
!
      case(1)

#ifdef TVD
        v(2:im1,:,:) = dxih *       (u(3:im ,:,:) - u(1:im2,:,:))
#else /* TVD */
        v(3:im2,:,:) = dxi  * (cl * (u(4:im1,:,:) - u(2:im3,:,:))              &
                             - cr * (u(5:im ,:,:) - u(1:im4,:,:)))
        v(2    ,:,:) = dxih *       (u(3    ,:,:) - u(1    ,:,:))
        v(  im1,:,:) = dxih *       (u(  im ,:,:) - u(  im1,:,:))
#endif /* TVD */
        v(1    ,:,:) = dxi  *       (u(2    ,:,:) - u(1    ,:,:))
        v(  im ,:,:) = dxi  *       (u(  im ,:,:) - u(  im1,:,:))

! derivative along the Y direction
!
      case(2)

#ifdef TVD
        v(:,2:jm1,:) = dyih *       (u(:,3:jm ,:) - u(:,1:jm2,:))
#else /* TVD */
        v(:,3:jm2,:) = dyi  * (cl * (u(:,4:jm1,:) - u(:,2:jm3,:))              &
                             - cr * (u(:,5:jm ,:) - u(:,1:jm4,:)))
        v(:,2    ,:) = dyih *       (u(:,3    ,:) - u(:,1    ,:))
        v(:,  jm1,:) = dyih *       (u(:,  jm ,:) - u(:,  jm2,:))
#endif /* TVD */
        v(:,1    ,:) = dyi  *       (u(:,2    ,:) - u(:,1    ,:))
        v(:,  jm ,:) = dyi  *       (u(:,  jm ,:) - u(:,  jm1,:))

! derivative along the Z direction
!
      case(3)

#ifdef R3D
#ifdef TVD
        v(:,:,2:km1) = dzih *       (u(:,:,3:km ) - u(:,:,1:km2))
#else /* TVD */
        v(:,:,3:km2) = dzi  * (cl * (u(:,:,4:km1) - u(:,:,2:km3))              &
                             - cr * (u(:,:,5:km ) - u(:,:,1:km4)))
        v(:,:,2    ) = dzih *       (u(:,:,3    ) - u(:,:,1    ))
        v(:,:,  km1) = dzih *       (u(:,:,  km ) - u(:,:,  km2))
#endif /* TVD */
        v(:,:,1    ) = dzi  *       (u(:,:,2    ) - u(:,:,1    ))
        v(:,:,  km ) = dzi  *       (u(:,:,  km ) - u(:,:,  km1))
#else /* R3D */
        v(:,:,:    ) = 0.0d0
#endif /* R3D */

      end select

    end if
#endif /* DETECT_DISCONTINUITY */

!-------------------------------------------------------------------------------
!
  end subroutine derivative_1st
!
!===============================================================================
!
! subroutine DERIVATIVE_2ND:
! -------------------------
!
!   Subroutine calculates the second order derivative of the input vector.
!
!   Arguments:
!
!     n - the direction of derivative
!     u - the input vector
!     v - the second derivative of the input vector
!
!===============================================================================
!
  subroutine derivative_2nd(n, im, jm, km, u, v)

! include external procedures and variables
!
    use mesh, only : dxi, dyi, dzi

! local variables are not implicit by default
!
    implicit none

! subroutine arguments
!
    integer                   , intent(in)  :: n, im, jm, km
    real, dimension(im,jm,km) , intent(in)  :: u
    real, dimension(im,jm,km) , intent(out) :: v

! local variables
!
    logical, save   :: first = .true.
    integer         :: im1, im2, jm1, jm2, km1, km2
#ifndef TVD
    integer         :: im3, im4, jm3, jm4, km3, km4
#endif /* !TVD */
    real   , save   :: dxi2, dyi2, dzi2

! coefficients for the 5th order derivative estimation
!
#ifndef TVD
    real, parameter :: f0 = - 5.0d0 /  2.0d0, f1 =   4.0d0 /  3.0d0            &
                     , f2 = - 1.0d0 / 12.0d0
#endif /* !TVD */
!
!-------------------------------------------------------------------------------
!
    if (first) then

! calculate temporary spatial steps
!
      dxi2 = dxi * dxi
      dyi2 = dyi * dyi
#ifdef R3D
      dzi2 = dzi * dzi
#endif /* R3D */

! reset the flag 'first'
!
      first = .false.

    end if

! prepare index limits
!
    im1 = im - 1
    im2 = im - 2
#ifndef TVD
    im3 = im - 3
    im4 = im - 4
#endif /* !TVD */
    jm1 = jm - 1
    jm2 = jm - 2
#ifndef TVD
    jm3 = jm - 3
    jm4 = jm - 4
#endif /* !TVD */
#ifdef R3D
    km1 = km - 1
    km2 = km - 2
#ifndef TVD
    km3 = km - 3
    km4 = km - 4
#endif /* !TVD */
#endif /* R3D */

! select the direction
!
    select case(n)

! derivative along the X direction
!
    case(1)

#ifdef TVD
      v(2:im1,:,:) = ((u(3:im ,:,:) + u(1:im2,:,:))                            &
                                                - 2.0d0 * u(2:im1,:,:)) * dxi2
#else /* TVD */
      v(3:im2,:,:) = (f0 *  u(3:im2,:,:)                                       &
                    + f1 * (u(4:im1,:,:) + u(2:im3,:,:))                       &
                    + f2 * (u(5:im ,:,:) + u(1:im4,:,:))) * dxi2
      v(2    ,:,:) = ((u(3    ,:,:) + u(1    ,:,:))                            &
                                                - 2.0d0 * u(2    ,:,:)) * dxi2
      v(  im1,:,:) = ((u(  im ,:,:) + u(  im2,:,:))                            &
                                                - 2.0d0 * u(  im1,:,:)) * dxi2
#endif /* TVD */
      v(1    ,:,:) = 0.0d0
      v(  im ,:,:) = 0.0d0

! derivative along the Y direction
!
    case(2)

#ifdef TVD
      v(:,2:jm1,:) = ((u(:,3:jm ,:) + u(:,1:jm2,:))                            &
                                                - 2.0d0 * u(:,2:jm1,:)) * dyi2
#else /* TVD */
      v(:,3:jm2,:) = (f0 *  u(:,3:jm2,:)                                       &
                    + f1 * (u(:,4:jm1,:) + u(:,2:jm3,:))                       &
                    + f2 * (u(:,5:jm ,:) + u(:,1:jm4,:))) * dyi2
      v(:,2    ,:) = ((u(:,3    ,:) + u(:,1    ,:))                            &
                                                - 2.0d0 * u(:,2    ,:)) * dyi2
      v(:,  jm1,:) = ((u(:,  jm ,:) + u(:,  jm2,:))                            &
                                                - 2.0d0 * u(:,  jm1,:)) * dyi2
#endif /* TVD */
      v(:,1    ,:) = 0.0d0
      v(:,  jm ,:) = 0.0d0

! derivative along the Z direction
!
      case(3)

#ifdef R3D
#ifdef TVD
      v(:,:,2:km1) = ((u(:,:,3:km ) + u(:,:,1:km2))                            &
                                                - 2.0d0 * u(:,:,2:km1)) * dzi2
#else /* TVD */
      v(:,:,3:km2) = (f0 *  u(:,:,3:km2)                                       &
                    + f1 * (u(:,:,4:km1) + u(:,:,2:km3))                       &
                    + f2 * (u(:,:,5:km ) + u(:,:,1:km4))) * dzi2
      v(:,:,2    ) = ((u(:,:,3    ) + u(:,:,1    ))                            &
                                                - 2.0d0 * u(:,:,2    )) * dzi2
      v(:,:,  km1) = ((u(:,:,  km ) + u(:,:,  km2))                            &
                                                - 2.0d0 * u(:,:,  km1)) * dzi2
#endif /* TVD */
      v(:,:,1    ) = 0.0d0
      v(:,:,  km ) = 0.0d0
#else /* R3D */
      v(:,:, :   ) = 0.0d0
#endif /* R3D */

    end select

!-------------------------------------------------------------------------------
!
  end subroutine derivative_2nd
!
!===============================================================================
!
! subroutine DIVERGENCE:
! ---------------------
!
!   Subroutine calculates the cell centered divergence of the input vector
!   field.
!
!===============================================================================
!
  subroutine divergence(im, jm, km, u, v, t)

! local variables are not implicit by default
!
    implicit none

! input and output variables
!
    integer                    , intent(in)  :: im, jm, km
    real, dimension(3,im,jm,km), intent(in)  :: u
    real, dimension(  im,jm,km), intent(out) :: v
    character(len=1)           , intent(in)  :: t

! local arrays
!
    real, dimension(  im,jm,km)                :: w
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! start accounting time for the divergence operator calculation
!
    call start_timer(idv)
#endif /* PROFILE */

! calculate contribution from the X derivative of the X component
!
    call derivative_1st(1, im, jm, km, u(inx,:,:,:), w(:,:,:), t)

! update the divergence array
!
    v(:,:,:) = w(:,:,:)

! calculate contribution from the Y derivative of the Y component
!
    call derivative_1st(2, im, jm, km, u(iny,:,:,:), w(:,:,:), t)

! update the divergence array
!
    v(:,:,:) = v(:,:,:) + w(:,:,:)

#ifdef R3D
! calculate contribution from the Z derivative of the Z component
!
    call derivative_1st(3, im, jm, km, u(inz,:,:,:), w(:,:,:), t)

! update the divergence array
!
    v(:,:,:) = v(:,:,:) + w(:,:,:)
#endif /* R3D */

#ifdef PROFILE
! stop accounting time for the divergence operator calculation
!
    call stop_timer(idv)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine divergence
!
!===============================================================================
!
! subroutine CURL:
! ---------------
!
!   Subroutine calculates the curl of the input vector field.
!
!   For CD methods:
!            input field is located at cell centers
!                    [Ax(i,j,k), Ay(i,j,k), Az(i,j,k)]
!            output field is located at cell centers
!                    [Bx(i,j,k), By(i,j,k), Bz(i,j,k)]
!
!   For CT methods:
!            input field is located at cell edges
!                    [Ax(i,j+1/2,k+1/2), Ay(i+1/2,j,k+1/2), Az(i+1/2,j+1/2,k)]
!            output field is located at cell interfaces
!                    [Bx(i+1/2,j,k), By(i,j+1/2,k), Bz(i,j,k+1/2)]
!
!      Bx = d/dy Az - d/dz Ay
!      By = d/dz Ax - d/dx Az
!      Bz = d/dx Ay - d/dy Ax
!
!   Note:
!
!      The input values are assumed to be cell integrals, therefore the high
!      order estimation of the first order derivative uses a interpolation
!      based on cell integrals providing the cell integral values of the output
!      vector field.
!
!===============================================================================
!
  subroutine curl(im, jm, km, a, b, t)

! local variables are not implicit by default
!
    implicit none

! input and output variables
!
    integer                    , intent(in)  :: im, jm, km
    real, dimension(3,im,jm,km), intent(in)  :: a
    real, dimension(3,im,jm,km), intent(out) :: b
    character(len=1)           , intent(in)  :: t

! local arrays
!
    real, dimension(  im,jm,km)              :: w
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! start accounting time for the curl operation
!
    call start_timer(icr)
#endif /* PROFILE */

! === calculate Bx component ===
!
! contribution from the Y derivative of Az
!
    call derivative_1st(2, im, jm, km, a(inz,:,:,:), w(:,:,:), t)

! update Bx
!
    b(inx,:,:,:) = w(:,:,:)

#ifdef R3D
! contribution from the Z derivative of Ay
!
    call derivative_1st(3, im, jm, km, a(iny,:,:,:), w(:,:,:), t)

! update Bx
!
    b(inx,:,:,:) = b(inx,:,:,:) - w(:,:,:)
#endif /* R3D */

! === calculate By component ===
!
#ifdef R3D
! contribution from the Z derivative of Ax
!
    call derivative_1st(3, im, jm, km, a(inx,:,:,:), w(:,:,:), t)

! update By
!
    b(iny,:,:,:) = w(:,:,:)

! contribution from the X derivative of Az
!
    call derivative_1st(1, im, jm, km, a(inz,:,:,:), w(:,:,:), t)

! update By
!
    b(iny,:,:,:) = b(iny,:,:,:) - w(:,:,:)
#else /* R3D */
! contribution from the X derivative of Az
!
    call derivative_1st(1, im, jm, km, a(inz,:,:,:), w(:,:,:), t)

! update By
!
    b(iny,:,:,:) = - w(:,:,:)
#endif /* R3D */

! === calculate Bz component ===
!
! contribution from the X derivative of Ay
!
    call derivative_1st(1, im, jm, km, a(iny,:,:,:), w(:,:,:), t)

! update Bz
!
    b(inz,:,:,:) = w(:,:,:)

! contribution from the Y derivative of Ax
!
    call derivative_1st(2, im, jm, km, a(inx,:,:,:), w(:,:,:), t)

! update Bz
!
    b(inz,:,:,:) = b(inz,:,:,:) - w(:,:,:)

#ifdef PROFILE
! stop accounting time for the curl operation
!
    call stop_timer(icr)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine curl
!
!===============================================================================
!
! subroutine LAPLACE:
! ------------------
!
!   Subroutine aplies the Laplace operator in the input scalar field.
!
!     B = Δ A = (d²/dx² + d²/dy² + d²/dz²) A
!
!   Note:
!
!      The input values are assumed to be cell integrals, therefore the high
!      order estimation of the first order derivative uses a interpolation
!      based on cell integrals providing the cell integral values of the output
!      vector field.
!
!   Arguments:
!
!     a - the input vector
!     b - the Laplacian of the input vector
!
!===============================================================================
!
  subroutine laplace(im, jm, km, a, b)

! local variables are not implicit by default
!
    implicit none

! subroutine arguments
!
    integer                  , intent(in)  :: im, jm, km
    real, dimension(im,jm,km), intent(in)  :: a
    real, dimension(im,jm,km), intent(out) :: b

! local arrays
!
    real, dimension(im,jm,km) :: w
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! start accounting time for the Laplace operator calculation
!
    call start_timer(ilp)
#endif /* PROFILE */

! calculate the second derivative along the X direction
!
    call derivative_2nd(1, im, jm, km, a(:,:,:), w(:,:,:))

! update the output field
!
    b(:,:,:) = w(:,:,:)

! calculate the second derivative along the Y direction
!
    call derivative_2nd(2, im, jm, km, a(:,:,:), w(:,:,:))

! update the output field
!
    b(:,:,:) = b(:,:,:) + w(:,:,:)

#ifdef R3D
! calculate the second derivative along the Z direction
!
    call derivative_2nd(3, im, jm, km, a(:,:,:), w(:,:,:))

! update the output field
!
    b(:,:,:) = b(:,:,:) + w(:,:,:)
#endif /* R3D */

#ifdef PROFILE
! stop accounting time for the Laplace operator calculation
!
    call stop_timer(ilp)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine laplace
!
!===============================================================================
!
! subroutine INTERPOLATE:
! ----------------------
!
!   Subroutine interpolates the input field between it cell center and cell
!   interface representations.
!
!   Arguments:
!
!     n - the direction of the interpolation
!     u - the input vector at selected location
!     v - the interpolation output vector
!     t - optional flag to specify from which location we interpolate
!         'f' - from the interface to center representation
!         'c' - from the center to interface representation
!
!===============================================================================
!
  subroutine interpolate(n, im, jm, km, u, v, t)

! local variables are not implicit by default
!
    implicit none

! subroutine arguments
!
    integer                   , intent(in)  :: n, im, jm, km
    real, dimension(im,jm,km) , intent(in)  :: u
    real, dimension(im,jm,km) , intent(out) :: v
    character(len=1)          , intent(in)  :: t

! local variables
!
    integer, save      :: imm1, jmm1, kmm1
#ifndef TVD
    integer, save      :: imm2, imm3, jmm2, jmm3, kmm2, kmm3
#endif /* !TVD */

#ifndef TVD
! coefficients for the 4th order interpolation
!
    real   , parameter :: cl = 7.0d+00 / 12.0d+00, cr  = 1.0d+00 / 12.0d+00
    real   , parameter :: fl = 9.0d+00 / 16.0d+00, fr  = 1.0d+00 / 16.0d+00

#ifdef DETECT_DISCONTINUITY
! coefficients for the second order interpolation used for the discontinuity
! detection
!
    real   , parameter :: f1 = 0.750d+00, f2 = 0.375d+00, f3 = 0.125d+00
#endif /* DETECT_DISCONTINUITY */
#endif /* !TVD */

! local variables
!
    integer :: i , j , k
#ifdef DETECT_DISCONTINUITY
    real    :: vl, vu
#endif /* DETECT_DISCONTINUITY */
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! start accounting time for the interpolate operator
!
    call start_timer(iin)
#endif /* PROFILE */

! prepare index limits during the first execution
!
    imm1 = im - 1
#ifndef TVD
    imm2 = im - 2
    imm3 = im - 3
#endif /* !TVD */
    jmm1 = jm - 1
#ifndef TVD
    jmm2 = jm - 2
    jmm3 = jm - 3
#endif /* !TVD */
#ifdef R3D
    kmm1 = km - 1
#ifndef TVD
    kmm2 = km - 2
    kmm3 = km - 3
#endif /* !TVD */
#endif /* R3D */

! check if the additional argument is present and if it is, check if it
! indicates to interpolate from the the interface values
!
    if (t == 'f') then

! select the direction
!
      select case(n)

! interpolate from the right X interface to the center
!
      case(1)

        do k = 1, km
          do j = 1, jm
#ifdef TVD
            v(2:im  ,j,k) = 0.5d+00 * (u(1:imm1,j,k) + u(2:im  ,j,k))
#else /* TVD */
            v(3:imm1,j,k) =      fl * (u(2:imm2,j,k) + u(3:imm1,j,k))          &
                            -    fr * (u(1:imm3,j,k) + u(4:im  ,j,k))
#ifdef DETECT_DISCONTINUITY
            do i = 3, imm1
              vl       = f1 * u(i-1,j,k) + f2 * u(i  ,j,k) - f3 * u(i-2,j,k)
              vu       = f1 * u(i  ,j,k) + f2 * u(i-1,j,k) - f3 * u(i+1,j,k)
              if ((vl - v(i,j,k)) * (vu - v(i,j,k)) > 0.0d+00) then
                v(i,j,k) = 0.5d+00 * (vl + vu)
              end if
            end do
#endif /* DETECT_DISCONTINUITY */
            v(im  ,j,k) = 0.5d+00 * (u(imm1,j,k) + u(im  ,j,k))
            v( 2  ,j,k) = 0.5d+00 * (u( 1  ,j,k) + u( 2  ,j,k))
#endif /* TVD */
            v( 1  ,j,k) = 1.5d+00 *  u( 1  ,j,k) - 0.5d+00 * u( 2  ,j,k)
          end do
        end do

! interpolate from the right Y interface to the center
!
      case(2)

        do k = 1, km
          do i = 1, im
#ifdef TVD
            v(i,2:jm  ,k) = 0.5d+00 * (u(i,1:jmm1,k) + u(i,2:jm  ,k))
#else /* TVD */
            v(i,3:jmm1,k) =      fl * (u(i,2:jmm2,k) + u(i,3:jmm1,k))          &
                            -    fr * (u(i,1:jmm3,k) + u(i,4:jm  ,k))
#ifdef DETECT_DISCONTINUITY
            do j = 3, jmm1
              vl       = f1 * u(i,j-1,k) + f2 * u(i,j  ,k) - f3 * u(i,j-2,k)
              vu       = f1 * u(i,j  ,k) + f2 * u(i,j-1,k) - f3 * u(i,j+1,k)
              if ((vl - v(i,j,k)) * (vu - v(i,j,k)) > 0.0d+00) then
                v(i,j,k) = 0.5d+00 * (vl + vu)
              end if
            end do
#endif /* DETECT_DISCONTINUITY */
            v(i,jm  ,k) = 0.5d+00 * (u(i,jmm1,k) + u(i,jm  ,k))
            v(i, 2  ,k) = 0.5d+00 * (u(i, 1  ,k) + u(i, 2  ,k))
#endif /* TVD */
            v(i, 1  ,k) = 1.5d+00 *  u(i, 1  ,k) - 0.5d+00 * u(i, 2  ,k)
          end do
        end do

! interpolate from the right Z interface to the center
!
      case(3)

#ifdef R3D
        do j = 1, jm
          do i = 1, im
#ifdef TVD
            v(i,j,2:km  ) = 0.5d+00 * (u(i,j,1:kmm1) + u(i,j,2:km  ))
#else /* TVD */
            v(i,j,3:kmm1) =      fl * (u(i,j,2:kmm2) + u(i,j,3:kmm1))          &
                            -    fr * (u(i,j,1:kmm3) + u(i,j,4:km  ))
#ifdef DETECT_DISCONTINUITY
            do k = 3, kmm1
              vl       = f1 * u(i,j,k-1) + f2 * u(i,j,k  ) - f3 * u(i,j,k-2)
              vu       = f1 * u(i,j,k  ) + f2 * u(i,j,k-1) - f3 * u(i,j,k+1)
              if ((vl - v(i,j,k)) * (vu - v(i,j,k)) > 0.0d+00) then
                v(i,j,k) = 0.5d+00 * (vl + vu)
              end if
            end do
#endif /* DETECT_DISCONTINUITY */
            v(i,j,km  ) = 0.5d+00 * (u(i,j,kmm1) + u(i,j,km  ))
            v(i,j, 2  ) = 0.5d+00 * (u(i,j, 1  ) + u(i,j, 2  ))
#endif /* TVD */
            v(i,j, 1  ) = 1.5d+00 *  u(i,j, 1  ) - 0.5d+00 * u(i,j, 2  )
          end do
        end do
#else /* R3D */
        v(:,:,:) = u(:,:,:)
#endif /* R3D */

      end select

    else

! select the direction
!
      select case(n)

! interpolate from the center to the right X interface
!
      case(1)

        do k = 1, km
          do j = 1, jm
#ifdef TVD
            v(1:imm1,j,k) = 0.5d+00 * (u(1:imm1,j,k) + u(2:im  ,j,k))
#else /* TVD */
            v(2:imm2,j,k) =      cl * (u(2:imm2,j,k) + u(3:imm1,j,k))          &
                            -    cr * (u(1:imm3,j,k) + u(4:im  ,j,k))
#ifdef DETECT_DISCONTINUITY
            do i = 2, imm2
              vl       = f1 * u(i  ,j,k) + f2 * u(i+1,j,k) - f3 * u(i-1,j,k)
              vu       = f1 * u(i+1,j,k) + f2 * u(i  ,j,k) - f3 * u(i+2,j,k)
              if ((vl - v(i,j,k)) * (vu - v(i,j,k)) > 0.0d+00) then
                v(i,j,k) = 0.5d+00 * (vl + vu)
              end if
            end do
#endif /* DETECT_DISCONTINUITY */
            v( 1  ,j,k) = 0.5d+00 * (u( 1  ,j,k) + u( 2  ,j,k))
            v(imm1,j,k) = 0.5d+00 * (u(imm1,j,k) + u(im  ,j,k))
#endif /* TVD */
            v(im  ,j,k) = 1.5d+00 *  u(im  ,j,k) - 0.5d+00 * u(imm1,j,k)
          end do
        end do

! interpolate from the center to the right Y interface
!
      case(2)

        do k = 1, km
          do i = 1, im
#ifdef TVD
            v(i,1:jmm1,k) = 0.5d+00 * (u(i,1:jmm1,k) + u(i,2:jm  ,k))
#else /* TVD */
            v(i,2:jmm2,k) =      cl * (u(i,2:jmm2,k) + u(i,3:jmm1,k))          &
                            -    cr * (u(i,1:jmm3,k) + u(i,4:jm  ,k))
#ifdef DETECT_DISCONTINUITY
            do j = 2, jmm2
              vl       = f1 * u(i,j  ,k) + f2 * u(i,j+1,k) - f3 * u(i,j-1,k)
              vu       = f1 * u(i,j+1,k) + f2 * u(i,j  ,k) - f3 * u(i,j+2,k)
              if ((vl - v(i,j,k)) * (vu - v(i,j,k)) > 0.0d+00) then
                v(i,j,k) = 0.5d+00 * (vl + vu)
              end if
            end do
#endif /* DETECT_DISCONTINUITY */
            v(i, 1  ,k) = 0.5d+00 * (u(i, 1  ,k) + u(i, 2  ,k))
            v(i,jmm1,k) = 0.5d+00 * (u(i,jmm1,k) + u(i,jm  ,k))
#endif /* TVD */
            v(i,jm  ,k) = 1.5d+00 *  u(i,jm  ,k) - 0.5d+00 * u(i,jmm1,k)
          end do
        end do

! interpolate from the center to the right Z interface
!
      case(3)

#ifdef R3D
        do j = 1, jm
          do i = 1, im
#ifdef TVD
            v(i,j,1:kmm1) = 0.5d+00 * (u(i,j,1:kmm1) + u(i,j,2:km  ))
#else /* TVD */
            v(i,j,2:kmm2) =      cl * (u(i,j,2:kmm2) + u(i,j,3:kmm1))          &
                            -    cr * (u(i,j,1:kmm3) + u(i,j,4:km  ))
#ifdef DETECT_DISCONTINUITY
            do k = 2, kmm2
              vl       = f1 * u(i,j,k  ) + f2 * u(i,j,k+1) - f3 * u(i,j,k-1)
              vu       = f1 * u(i,j,k+1) + f2 * u(i,j,k  ) - f3 * u(i,j,k+2)
              if ((vl - v(i,j,k)) * (vu - v(i,j,k)) > 0.0d+00) then
                v(i,j,k) = 0.5d+00 * (vl + vu)
              end if
            end do
#endif /* DETECT_DISCONTINUITY */
            v(i,j, 1  ) = 0.5d+00 * (u(i,j, 1  ) + u(i,j, 2  ))
            v(i,j,kmm1) = 0.5d+00 * (u(i,j,kmm1) + u(i,j,km  ))
#endif /* TVD */
            v(i,j,km  ) = 1.5d+00 *  u(i,j,km  ) - 0.5d+00 * u(i,j,kmm1)
          end do
        end do
#else /* R3D */
        v(:,:,:) = u(:,:,:)
#endif /* R3D */

      end select

    end if

#ifdef PROFILE
! stop accounting time for the interpolate operator
!
    call stop_timer(iin)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine interpolate
!
!===============================================================================
!
! subroutine ATOB:
! ---------------
!
!   Subroutine calculates the staggered magnetic field components vector
!   potential.
!
!   Note:
!
!      The input values are assumed to be cell integrals, therefore first we use
!      polynomial interpolation to calculate the point values at the cell
!      interface from cell integrals and then we use high order estimation for
!      point values to obtain the derivatives. This provides point values of
!      output vector field.
!
!===============================================================================
!
  subroutine atob(im, jm, km, a, b, t)

! local variables are not implicit by default
!
    implicit none

! input and output variables
!
    integer                    , intent(in)  :: im, jm, km
    real, dimension(3,im,jm,km), intent(in)  :: a
    real, dimension(3,im,jm,km), intent(out) :: b
    character(len=1)           , intent(in)  :: t

! local arrays
!
    real, dimension(im,jm,km) :: ai, w
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! start accounting time for the interface magnetic field calculation
!
    call start_timer(iab)
#endif /* PROFILE */

! === calculate Bx component ===
!
! interpolate Az along the X direction
!
    call interpolate(1, im, jm, km, a(inz,:,:,:), ai(:,:,:), t)

! contribution from the Y derivative of Az
!
    call derivative_1st(2, im, jm, km, ai(:,:,:), w(:,:,:), t)

! update Bx
!
    b(inx,:,:,:) = w(:,:,:)

#ifdef R3D
! interpolate Ay along the X direction
!
    call interpolate(1, im, jm, km, a(iny,:,:,:), ai(:,:,:), t)

! contribution from the Z derivative of Ay
!
    call derivative_1st(3, im, jm, km, ai(:,:,:), w(:,:,:), t)

! update Bx
!
    b(inx,:,:,:) = b(inx,:,:,:) - w(:,:,:)
#endif /* R3D */

! === calculate By component ===
!
#ifdef R3D
! interpolate Ax along the Y direction
!
    call interpolate(2, im, jm, km, a(inx,:,:,:), ai(:,:,:), t)

! contribution from the Z derivative of Ax
!
    call derivative_1st(3, im, jm, km, ai(:,:,:), w(:,:,:), t)

! update By
!
    b(iny,:,:,:) = w(:,:,:)

! interpolate Az along the Y direction
!
    call interpolate(2, im, jm, km, a(inz,:,:,:), ai(:,:,:), t)

! contribution from the X derivative of Az
!
    call derivative_1st(1, im, jm, km, ai(:,:,:), w(:,:,:), t)

! update By
!
    b(iny,:,:,:) = b(iny,:,:,:) - w(:,:,:)
#else /* R3D */

! interpolate Az along the Y direction
!
    call interpolate(2, im, jm, km, a(inz,:,:,:), ai(:,:,:), t)

! contribution from the X derivative of Az
!
    call derivative_1st(1, im, jm, km, ai(:,:,:), w(:,:,:), t)

! update By
!
    b(iny,:,:,:) = - w(:,:,:)
#endif /* R3D */

! === calculate Bz component ===
!
#ifdef R2D
! reset the Z magnetic field component
!
    b(inz,:,:,:) = 0.0d0
#else /* R2D */

! interpolate Ay along the Z direction
!
    call interpolate(3, im, jm, km, a(iny,:,:,:), ai(:,:,:), t)

! contribution from the X derivative of Ay
!
    call derivative_1st(1, im, jm, km, ai(:,:,:), w(:,:,:), t)

! update Bz
!
    b(inz,:,:,:) = w(:,:,:)

! interpolate Ax along the Z direction
!
    call interpolate(3, im, jm, km, a(inx,:,:,:), ai(:,:,:), t)

! contribution from the Y derivative of Ax
!
    call derivative_1st(2, im, jm, km, ai(:,:,:), w(:,:,:), t)

! update Bz
!
    b(inz,:,:,:) = b(inz,:,:,:) - w(:,:,:)
#endif /* R2D */

#ifdef PROFILE
! stop accounting time for the interface magnetic field calculation
!
    call stop_timer(iab)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine atob
!
!===============================================================================
!
! subroutine ATOJ:
! ---------------
!
!   Subroutine calculates the current density directly from the vector
!   potential. In this way the current density is better localized than when it
!   is calculated from magnetic field.
!
!   For CT methods:
!            input field is located at cell centers:
!                    [Ax(i,j+1/2,k+1/2), Ay(i+1/2,j,k+1/2), Az(i+1/2,j+1/2,k)]
!            output field is located at cell centers
!                    [Jx(i,j+1/2,k+1/2), Jy(i+1/2,j,k+1/2), Jz(i+1/2,j+1/2,k)]
!
!   For CD methods:
!            input field is located at cell centers:
!                    [Ax(i,j,k), Ay(i,j,k), Az(i,j,k)]
!            output field is located at cell centers
!                    [Jx(i,j,k), Jy(i,j,k), Jz(i,j,k)]
!
!   The applied operator:
!            J = - nabla(A) + grad(div(A)) (in 3D)
!            J = - nabla(A)                (in 2D)
!
!===============================================================================
!
  subroutine atoj(im, jm, km, a, j)

! include external procedures and variables
!
    use mesh  , only : dxi, dyi, dzi, dxih, dyih, dzih

! local variables are not implicit by default
!
    implicit none

! input and output variables
!
    integer                    , intent(in)  :: im, jm, km
    real, dimension(3,im,jm,km), intent(in)  :: a
    real, dimension(3,im,jm,km), intent(out) :: j

! local index variables
!
#ifndef R2D
    integer :: n
    integer :: im1, im2, jm1, jm2, km1, km2
#endif /* !R2D */

! local arrays
!
    real, dimension(im,jm,km) :: da, ga
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! start accounting time for the calculation of J from A
!
    call start_timer(iaj)
#endif /* PROFILE */

#ifndef R2D
! calculate temporary indices
!
    im1 = im - 1
    im2 = im - 2
    jm1 = jm - 1
    jm2 = jm - 2
#ifdef R3D
    km1 = km - 1
    km2 = km - 2
#endif /* R3D */
#endif /* !R2D */

#ifdef R2D
! reset the current density
!
    j(inx,:,:,:) = 0.0d0
    j(iny,:,:,:) = 0.0d0

! calculate the Laplacian of the Z component of the current density
!
    call laplace(im, jm, km, a(inz,:,:,:), da(:,:,:))

! calculate the Z component of the current density
!
    j(inz,:,:,:) = - da(:,:,:)

#else /* R2D */
    do n = inx, inz

! calculate the Laplacian of the current component of vector potential
!
      call laplace(im, jm, km, a(n,:,:,:), da(:,:,:))

! calculate the current component of the current density
!
      j(n,:,:,:) = - da(:,:,:)

    end do ! n = inx, inz

#ifndef DETECT_DISCONTINUITY
#ifdef CT
! calculate the divergence of vector potential
!
    call divergence(im, jm, km, a(inx:inz,:,:,:), da(:,:,:), 'b')

! update the current density with the gradient of the divergence of A
!
    do n = inx, inz
      call derivative_1st(n, im, jm, km, da(:,:,:), ga(:,:,:), 'f')
      j(n,:,:,:) = j(n,:,:,:) + ga(:,:,:)
    end do
#endif /* CT */
#ifdef CD
! calculate the divergence of vector potential
!
    call divergence(im, jm, km, a(inx:inz,:,:,:), da(:,:,:), 'c')

! update the current density with the gradient of the divergence of A
!
    j(inx,2:im1,:,:) = j(inx,2:im1,:,:) + (da(3:im,:,:) - da(1:im2,:,:)) * dxih
    j(inx,1    ,:,:) = j(inx,1    ,:,:) + (da(2   ,:,:) - da(1    ,:,:)) * dxi
    j(inx,  im ,:,:) = j(inx,  im ,:,:) + (da(  im,:,:) - da(  im1,:,:)) * dxi
    j(iny,:,2:jm1,:) = j(iny,:,2:jm1,:) + (da(:,3:jm,:) - da(:,1:jm2,:)) * dyih
    j(iny,:,1    ,:) = j(iny,:,1    ,:) + (da(:,2   ,:) - da(:,1    ,:)) * dyi
    j(iny,:,  jm ,:) = j(iny,:,  jm ,:) + (da(:,  jm,:) - da(:,  jm1,:)) * dyi
#ifdef R3D
    j(inz,:,:,2:km1) = j(inz,:,:,2:km1) + (da(:,:,3:km) - da(:,:,1:km2)) * dzih
    j(inz,:,:,1    ) = j(inz,:,:,1    ) + (da(:,:,2   ) - da(:,:,1    )) * dzi
    j(inz,:,:,  km ) = j(inz,:,:,  km ) + (da(:,:,  km) - da(:,:,  km1)) * dzi
#endif /* R3D*/
#endif /* CD */
#endif /* !DETECT_DISCONTINUITY */
#endif /* R2D */

#ifdef PROFILE
! stop accounting time
!
    call stop_timer(iaj)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine atoj

!===============================================================================
!
end module operators
